"""Naive surrogates."""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING, ClassVar

from attr import define, field

from baybe.surrogates.base import IndependentGaussianSurrogate
from baybe.surrogates.utils import batchify_mean_var_prediction

if TYPE_CHECKING:
    from torch import Tensor


@define
class MeanPredictionSurrogate(IndependentGaussianSurrogate):
    """A trivial surrogate model.

    It provides the average value of the training targets
    as posterior mean and a (data-independent) constant posterior variance.
    """

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    _model: float | None = field(init=False, default=None, eq=False)
    """The estimated posterior mean value of the training targets."""

    @batchify_mean_var_prediction
    def _estimate_moments(
        self, candidates_comp_scaled: Tensor, /
    ) -> tuple[Tensor, Tensor]:
        # See base class.

        import torch

        # TODO: use target value bounds for covariance scaling when explicitly provided
        mean = self._model * torch.ones([len(candidates_comp_scaled)])
        var = torch.ones(len(candidates_comp_scaled))
        return mean, var

    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        # See base class.
        self._model = train_y.mean().item()


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Optional polars import."""

from baybe.exceptions import OptionalImportError

try:
    import polars
except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Polars functionality is unavailable because 'polars' is not installed. "
        "Consider installing BayBE with 'polars' dependency, e.g. via "
        "`pip install baybe[polars]`."
    ) from ex

__all__ = [
    "polars",
]



"""Tests for memory utilities."""

from baybe.utils.memory import bytes_to_human_readable


def test_memory_human_readable_conversion():
    """The memory conversion to human readable format is correct."""
    assert bytes_to_human_readable(1024) == (1.0, "KB")
    assert bytes_to_human_readable(1024**2) == (1.0, "MB")
    assert bytes_to_human_readable(4.3 * 1024**4) == (4.3, "TB")



"""Hypothesis strategies for kernels."""

from enum import Enum

import hypothesis.strategies as st

from baybe.kernels.basic import (
    LinearKernel,
    MaternKernel,
    PeriodicKernel,
    PiecewisePolynomialKernel,
    PolynomialKernel,
    RBFKernel,
    RFFKernel,
    RQKernel,
)
from baybe.kernels.composite import AdditiveKernel, ProductKernel, ScaleKernel

from ..hypothesis_strategies.basic import positive_finite_floats
from ..hypothesis_strategies.priors import priors


class KernelType(Enum):
    """Taxonomy of kernel types."""

    SINGLE = "SINGLE"
    ADDITIVE = "ADDITIVE"
    PRODUCT = "PRODUCT"


linear_kernels = st.builds(
    LinearKernel,
    variance_prior=st.one_of(st.none(), priors()),
    variance_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates linear kernels."""

matern_kernels = st.builds(
    MaternKernel,
    nu=st.sampled_from((0.5, 1.5, 2.5)),
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates Matern kernels."""

periodic_kernels = st.builds(
    PeriodicKernel,
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
    period_length_prior=st.one_of(st.none(), priors()),
    period_length_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates periodic kernels."""

piecewise_polynomial_kernels = st.builds(
    PiecewisePolynomialKernel,
    q=st.integers(min_value=0, max_value=3),
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates piecewise polynomial kernels."""

polynomial_kernels = st.builds(
    PolynomialKernel,
    power=st.integers(min_value=0),
    offset_prior=st.one_of(st.none(), priors()),
    offset_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates polynomial kernels."""

rbf_kernels = st.builds(
    RBFKernel,
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates radial basis function (RBF) kernels."""

rff_kernels = st.builds(
    RFFKernel,
    num_samples=st.integers(min_value=1),
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates random Fourier features (RFF) kernels."""

rq_kernels = st.builds(
    RQKernel,
    lengthscale_prior=st.one_of(st.none(), priors()),
    lengthscale_initial_value=st.one_of(st.none(), positive_finite_floats()),
)
"""A strategy that generates rational quadratic (RQ) kernels."""

base_kernels = st.one_of(
    [
        matern_kernels,  # on top because it is the default for many use cases
        linear_kernels,
        rbf_kernels,
        rq_kernels,
        rff_kernels,
        piecewise_polynomial_kernels,
        polynomial_kernels,
        periodic_kernels,
    ]
)
"""A strategy that generates base kernels to be used within more complex kernels."""


@st.composite
def single_kernels(draw: st.DrawFn):
    """Generate single kernels (i.e., without kernel arithmetic, except scaling)."""
    base_kernel = draw(base_kernels)
    add_scale = draw(st.booleans())
    if add_scale:
        return ScaleKernel(
            base_kernel=base_kernel,
            outputscale_prior=draw(st.one_of(st.none(), priors())),
            outputscale_initial_value=draw(
                st.one_of(st.none(), positive_finite_floats()),
            ),
        )
    else:
        return base_kernel


@st.composite
def kernels(draw: st.DrawFn):
    """Generate :class:`baybe.kernels.base.Kernel`."""
    kernel_type = draw(st.sampled_from(KernelType))

    if kernel_type is KernelType.SINGLE:
        return draw(single_kernels())

    base_kernels = draw(st.lists(single_kernels(), min_size=2))
    if kernel_type is KernelType.ADDITIVE:
        return AdditiveKernel(base_kernels)
    if kernel_type is KernelType.PRODUCT:
        return ProductKernel(base_kernels)



"""Collection of small basic utilities."""

import functools
import inspect
from collections.abc import Callable, Collection, Iterable, Sequence
from dataclasses import dataclass
from typing import Any, TypeVar

from attrs import asdict, has

from baybe.exceptions import UnidentifiedSubclassError, UnmatchedAttributeError

_C = TypeVar("_C", bound=type)
_T = TypeVar("_T")
_U = TypeVar("_U")


@dataclass(frozen=True, repr=False)
class Dummy:
    """Placeholder element for array-like data types.

    Useful e.g. for detecting duplicates in constraints.
    """

    def __repr__(self):
        """Return a representation of the placeholder."""
        return "<dummy>"


def get_subclasses(cls: _C, recursive: bool = True, abstract: bool = False) -> list[_C]:
    """Return a list of subclasses for the given class.

    Args:
        cls: The base class to retrieve subclasses for.
        recursive: If ``True``, indirect subclasses (i.e. subclasses of subclasses)
            are included.
        abstract: If ``True``, abstract subclasses are included.

    Returns:
        A list of subclasses for the given class.
    """
    from baybe.utils.boolean import is_abstract

    subclasses = []
    for subclass in cls.__subclasses__():
        # Append direct subclass only if it is not abstract
        if abstract or not is_abstract(subclass):
            subclasses.append(subclass)

        # If requested, add indirect subclasses
        if recursive:
            subclasses.extend(get_subclasses(subclass, abstract=abstract))

    return subclasses


def get_baseclasses(
    cls: type,
    recursive: bool = True,
    abstract: bool = False,
) -> list[type]:
    """Return a list of base classes for the given class.

    Args:
        cls: The class to retrieve base classes for.
        recursive: If ``True``, indirect base classes (i.e., base classes of base
            classes) are included.
        abstract: If `True`, abstract base classes are included.

    Returns:
        A list of base classes for the given class.
    """
    from baybe.utils.boolean import is_abstract

    classes = []

    for baseclass in cls.__bases__:
        if baseclass not in classes:
            if abstract or not is_abstract(baseclass):
                classes.append(baseclass)

            if recursive:
                classes.extend(get_baseclasses(baseclass, abstract=abstract))

    return classes


def hilberts_factory(factory: Callable[..., _T]) -> Iterable[_T]:
    """Provide an infinite stream of the factory's products."""
    while True:
        yield factory()


def group_duplicate_values(dictionary: dict[_T, _U]) -> dict[_U, list[_T]]:
    """Identify groups of keys that have the same value.

    Args:
        dictionary: The dictionary to screen for duplicate values.

    Returns:
        A dictionary whose keys are a subset of values of the input dictionary,
        and whose values are lists that group original keys holding the same value.

    Example:
        >>> group_duplicate_values({"A": 1, "B": 2, "C": 1, "D": 3})
        {1: ['A', 'C']}
    """
    group: dict[_U, list[_T]] = {}
    for key, value in dictionary.items():
        group.setdefault(value, []).append(key)
    return {k: v for k, v in group.items() if len(v) > 1}


def to_tuple(x: Sequence) -> tuple:
    """Convert any sequence into a tuple.

    This wrapper is only used to avoid mypy warnings for attrs converters:
    * https://github.com/python/mypy/issues/8417
    * https://github.com/python/mypy/issues/8389
    * https://github.com/python/mypy/issues/5313
    """
    # TODO: Remove wrapper once mypy support is there
    return tuple(x)


def match_attributes(
    object: Any,
    callable_: Callable,
    /,
    strict: bool = True,
    ignore: Collection[str] = (),
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Find the attributes of an object that match with a given callable signature.

    Parameters appearing in the callable signature that have no match with the given
    object attributes are ignored.

    Args:
        object: The object whose attributes are to be matched.
        callable_: The callable against whose signature the attributes are to be
            matched.
        strict: If ``True``, an error is raised when the object has attributes that
            are not found in the callable signature (see also ``ignore``).
            If ``False``, these attributes are returned separately.
        ignore: A collection of attributes names that are to be ignored during matching.

    Raises:
        ValueError: If applied to a non-attrs object.
        UnmatchedAttributeError: In strict mode, if not all attributes can be matched.

    Returns:
        * A dictionary mapping the matched attribute names to their values.
        * A dictionary mapping the unmatched attribute names to their values.
    """
    if not has(object.__class__):
        raise ValueError(
            f"'{match_attributes.__name__}' only works with attrs objects."
        )
    # Get attribute/parameter sets
    set_object = set(asdict(object)) - set(ignore)
    set_callable = set(inspect.signature(callable_).parameters)

    # Match
    in_signature = set_object.intersection(set_callable)
    not_in_signature = set_object - set_callable
    if strict and not_in_signature:
        raise UnmatchedAttributeError(
            f"The following attributes cannot be matched: {not_in_signature}."
        )

    # Collect attributes for both sets
    attrs_in_signature = {p: getattr(object, p) for p in in_signature}
    attrs_not_in_signature = {p: getattr(object, p) for p in not_in_signature}

    return attrs_in_signature, attrs_not_in_signature


class classproperty:
    """A decorator to make class properties.

    A class property combines the characteristics of @property and @classmethod. The
    simple chaining of these two existing decorators is deprecated
    (https://docs.python.org/3.11/whatsnew/3.11.html#language-builtins) and causes mypy
    issues.
    """

    def __init__(self, fn: Callable) -> None:
        self.fn = fn

    def __get__(self, _, cl: type):
        return self.fn(cl)


def refers_to(cls: type, name_or_abbr: str, /) -> bool:
    """Check if the given name or abbreviation refers to the specified class."""
    return name_or_abbr in (
        (cls.__name__, cls.abbreviation)
        if hasattr(cls, "abbreviation")
        else (cls.__name__,)
    )


def find_subclass(base: type, name_or_abbr: str, /):
    """Retrieve a specific subclass of a base class via its name or abbreviation."""
    try:
        return next(cl for cl in get_subclasses(base) if refers_to(cl, name_or_abbr))
    except StopIteration:
        raise UnidentifiedSubclassError(
            f"The class name or abbreviation '{name_or_abbr}' does not refer to any "
            f"of the subclasses of '{base.__name__}'."
        )


def register_hooks(
    target: Callable,
    pre_hooks: Sequence[Callable] | None = None,
    post_hooks: Sequence[Callable] | None = None,
) -> Callable:
    """Register custom hooks with a given target callable.

    The provided hooks need to be "compatible" with the target in the sense that their
    signatures can be aligned:

    * The hook signature may only contain parameters that also exist in the target
      callable (<-- basic requirement).
    * However, parameters that are not needed by the hook can be omitted from
      its signature. This requires that the parameters of both signatures can be matched
      via their names. For simplicity, it is thus assumed that the hook has no
      positional-only arguments.
    * If an annotation is provided for a hook parameter, it must match its
      target counterpart (<-- safety mechanism to prevent unintended argument use).
      An exception is when the target parameter has no annotation, in which case
      the hook annotation is unrestricted. This is particularly useful when registering
      hooks with methods, since it offers the possibility to annotate the "self"
      parameter bound to the method-carrying object, which is typically not annotated
      in the target callable.

    Args:
        target: The callable to which the hooks are to be attached.
        pre_hooks: Hooks to be executed before calling the target.
        post_hooks: Hooks to be executed after calling the target.

    Returns:
        The wrapped callable with the hooks attached.

    Raises:
        TypeError: If any hook has positional-only arguments.
        TypeError: If any hook expects parameters that are not present in the target.
        TypeError: If any hook has a non-empty parameter annotation that does not
            match with the corresponding annotation of the target.
    """
    # Defaults
    pre_hooks = pre_hooks or []
    post_hooks = post_hooks or []

    target_signature = inspect.signature(target, eval_str=True).parameters

    # Validate hook signatures
    for hook in [*pre_hooks, *post_hooks]:
        hook_signature = inspect.signature(hook, eval_str=True).parameters

        if any(
            p.kind is inspect.Parameter.POSITIONAL_ONLY for p in hook_signature.values()
        ):
            raise TypeError("The provided hooks cannot have positional-only arguments.")

        if unrecognized := (set(hook_signature) - set(target_signature)):
            raise TypeError(
                f"The parameters expected by the hook '{hook.__name__}' must be a "
                f"subset of the parameter of the target callable '{target.__name__}'. "
                f"Unrecognized hook parameters: {unrecognized}."
            )

        for name, hook_param in hook_signature.items():
            target_param = target_signature[name]

            # If target parameter is not annotated, the hook annotation is unrestricted
            if (t_hint := target_param.annotation) is inspect.Parameter.empty:
                continue

            # If target parameter is annotated, the hook annotation must be compatible,
            # i.e., be identical or empty
            if ((h_hint := hook_param.annotation) != t_hint) and (
                h_hint is not inspect.Parameter.empty
            ):
                raise TypeError(
                    f"The type annotation for '{name}' is not consistent between "
                    f"the given hook '{hook.__name__}' and the target callable "
                    f"'{target.__name__}'. Given: {h_hint}. Expected: {t_hint}."
                )

    def pass_args(hook: Callable, *args, **kwargs) -> None:
        """Call the hook with its requested subset of arguments."""
        hook_signature = inspect.signature(hook, eval_str=True).parameters
        matched_args = dict(zip(target_signature, args))
        matched_kwargs = {
            p: kwargs.get(p, target_signature[p].default)
            for p in hook_signature
            if p not in matched_args
        }
        passed_kwargs = {p: (matched_args | matched_kwargs)[p] for p in hook_signature}
        hook(**passed_kwargs)

    @functools.wraps(target)
    def wraps(*args, **kwargs):
        for hook in pre_hooks:
            pass_args(hook, *args, **kwargs)
        result = target(*args, **kwargs)
        for hook in post_hooks:
            pass_args(hook, *args, **kwargs)
        return result

    return wraps



"""Base classes for all objectives."""

import gc
from abc import ABC, abstractmethod

import pandas as pd
from attrs import define
from cattrs import override

from baybe.objectives.deprecation import structure_objective
from baybe.serialization.core import (
    converter,
    unstructure_base,
)
from baybe.serialization.mixin import SerialMixin
from baybe.targets.base import Target

# TODO: Reactive slots in all classes once cached_property is supported:
#   https://github.com/python-attrs/attrs/issues/164


@define(frozen=True, slots=False)
class Objective(ABC, SerialMixin):
    """Abstract base class for all objectives."""

    @property
    @abstractmethod
    def targets(self) -> tuple[Target, ...]:
        """The targets included in the objective."""

    @abstractmethod
    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Transform target values from experimental to computational representation.

        Args:
            data: The data to be transformed. Must contain columns for all targets
                but can contain additional columns.

        Returns:
            A new dataframe with the targets in computational representation.
        """


def to_objective(x: Target | Objective, /) -> Objective:
    """Convert a target into an objective (with objective passthrough)."""
    return x if isinstance(x, Objective) else x.to_objective()


# Register de-/serialization hooks
converter.register_structure_hook(Objective, structure_objective)
converter.register_unstructure_hook(
    Objective,
    lambda x: unstructure_base(
        x,
        overrides=dict(
            _target=override(rename="target"),
            _targets=override(rename="targets"),
        ),
    ),
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Utilities for acquisition functions."""

from baybe.acquisition.base import AcquisitionFunction


def str_to_acqf(name: str, /) -> AcquisitionFunction:
    """Create an ACQF object from a given ACQF name."""
    return AcquisitionFunction.from_dict({"type": name})


def convert_acqf(acqf: AcquisitionFunction | str, /) -> AcquisitionFunction:
    """Convert an ACQF name into an ACQF object (with ACQF object passthrough)."""
    return acqf if isinstance(acqf, AcquisitionFunction) else str_to_acqf(acqf)



"""Hypothesis strategies for priors."""

import hypothesis.strategies as st

from baybe.priors import (
    BetaPrior,
    GammaPrior,
    HalfCauchyPrior,
    HalfNormalPrior,
    LogNormalPrior,
    NormalPrior,
    SmoothedBoxPrior,
)

from .basic import finite_floats
from .utils import intervals

gamma_priors = st.builds(
    GammaPrior,
    finite_floats(min_value=0.0, exclude_min=True),
    finite_floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Gamma priors."""

half_cauchy_priors = st.builds(
    HalfCauchyPrior,
    finite_floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Half-Cauchy priors."""

normal_priors = st.builds(
    NormalPrior,
    finite_floats(),
    finite_floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Normal priors."""

half_normal_priors = st.builds(
    HalfNormalPrior,
    finite_floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Half-Normal priors."""

log_normal_priors = st.builds(
    LogNormalPrior,
    finite_floats(),
    finite_floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Log-Normal priors."""

beta_priors = st.builds(
    BetaPrior,
    st.floats(min_value=0.0, exclude_min=True),
    st.floats(min_value=0.0, exclude_min=True),
)
"""A strategy that generates Beta priors."""


@st.composite
def _smoothed_box_priors(draw: st.DrawFn):
    """A strategy that generates Smoothed-Box priors."""
    interval = draw(intervals(exclude_half_bounded=True, exclude_fully_unbounded=True))
    sigma = draw(
        finite_floats(min_value=0.0, exclude_min=True),
    )

    return SmoothedBoxPrior(*interval.to_tuple(), sigma)


smoothed_box_priors = _smoothed_box_priors()


def priors(gpytorch_only: bool = True):
    """A strategy that generates priors."""
    prior_choices = [
        gamma_priors,
        half_cauchy_priors,
        half_normal_priors,
        log_normal_priors,
        normal_priors,
        smoothed_box_priors,
    ]
    if not gpytorch_only:
        prior_choices.append(beta_priors)
    return st.one_of(prior_choices)



"""Tests for basic input-output and iterative loop."""

import numpy as np
import pandas as pd
import pytest

from baybe.parameters import NumericalDiscreteParameter
from baybe.recommenders import BotorchRecommender
from baybe.searchspace import SearchSpace
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import add_fake_measurements

# List of tests that are expected to fail (still missing implementation etc)
param_xfails = []
target_xfails = []


@pytest.mark.parametrize(
    "bad_val",
    [1337, np.nan, "asd"],
    ids=["not_within_tol", "nan", "string_instead_float"],
)
def test_bad_parameter_input_value(campaign, good_reference_values, bad_val, request):
    """Test attempting to read in an invalid parameter value."""
    if request.node.callspec.id in param_xfails:
        pytest.xfail()

    rec = campaign.recommend(batch_size=3)
    add_fake_measurements(
        rec,
        campaign.targets,
        good_reference_values=good_reference_values,
    )

    # Add an invalid value
    rec.Num_disc_1.iloc[0] = bad_val
    with pytest.raises((ValueError, TypeError)):
        campaign.add_measurements(rec)


@pytest.mark.parametrize(
    "bad_val",
    [np.nan, "asd"],
    ids=["nan", "string_instead_float"],
)
def test_bad_target_input_value(campaign, good_reference_values, bad_val, request):
    """Test attempting to read in an invalid target value."""
    if request.node.callspec.id in target_xfails:
        pytest.xfail()

    rec = campaign.recommend(batch_size=3)
    add_fake_measurements(
        rec,
        campaign.targets,
        good_reference_values=good_reference_values,
    )

    # Add an invalid value
    rec.Target_max.iloc[0] = bad_val
    with pytest.raises((ValueError, TypeError)):
        campaign.add_measurements(rec)


@pytest.mark.parametrize("n_values", [5, 10, 20])
@pytest.mark.parametrize("n_parameters", [3, 5, 10])
def test_recommendation_is_not_ordered(n_values, n_parameters):
    """Test whether recommendations are unintentionally sorted.

    This is to ensure that they are not reordered according to the order they appear
    in the search space. To this end we create an example where the search space
    parameter entries are monotonically increasing and a target variable in max mode
    comprises the sum of the parameters. We feed the first and last point as training
    data. This should lead to a recommendation order that is very different from the
    original search space order (albeit not perfectly anti-ordered).
    """
    # Set up custom df with entries monotonically increasing
    values = list(range(n_values))
    df = pd.DataFrame({f"p{k+1}": values for k in range(n_parameters)})
    searchspace = SearchSpace.from_dataframe(
        df,
        parameters=[
            NumericalDiscreteParameter(name=f"p{k+1}", values=values)
            for k in range(n_parameters)
        ],
    )
    objective = NumericalTarget(name="t", mode="MAX").to_objective()
    recommender = BotorchRecommender()

    # Add first and last point as measurement, target value is the sum of all parameters
    measurements = df.iloc[[0, -1], :]
    measurements = measurements.assign(t=measurements.sum(axis=1))

    # Get recommendations and assert that they are not in ascending order regarding
    # their target value, as would be in the search space.
    rec = recommender.recommend(5, searchspace, objective, measurements)
    rec = rec.assign(t=rec.sum(axis=1))
    assert not rec["t"].is_monotonic_increasing, rec["t"].values



"""Gaussian process surrogates."""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING, ClassVar

from attrs import define, field
from attrs.validators import instance_of

from baybe.parameters.base import Parameter
from baybe.searchspace.core import SearchSpace
from baybe.surrogates.base import Surrogate
from baybe.surrogates.gaussian_process.kernel_factory import (
    KernelFactory,
    to_kernel_factory,
)
from baybe.surrogates.gaussian_process.presets import (
    GaussianProcessPreset,
    make_gp_from_preset,
)
from baybe.surrogates.gaussian_process.presets.default import (
    DefaultKernelFactory,
    _default_noise_factory,
)
from baybe.utils.plotting import to_string

if TYPE_CHECKING:
    from botorch.models.model import Model
    from botorch.models.transforms.input import InputTransform
    from botorch.models.transforms.outcome import OutcomeTransform
    from botorch.posteriors import Posterior
    from torch import Tensor


@define
class _ModelContext:
    """Model context for :class:`GaussianProcessSurrogate`."""

    searchspace: SearchSpace = field(validator=instance_of(SearchSpace))
    """The search space the model is trained on."""

    @property
    def task_idx(self) -> int | None:
        """The computational column index of the task parameter, if available."""
        return self.searchspace.task_idx

    @property
    def is_multitask(self) -> bool:
        """Indicates if model is to be operated in a multi-task context."""
        return self.n_task_dimensions > 0

    @property
    def n_task_dimensions(self) -> int:
        """The number of task dimensions."""
        # TODO: Generalize to multiple task parameters
        return 1 if self.task_idx is not None else 0

    @property
    def n_tasks(self) -> int:
        """The number of tasks."""
        return self.searchspace.n_tasks

    @property
    def parameter_bounds(self) -> Tensor:
        """Get the search space parameter bounds in BoTorch Format."""
        import torch

        return torch.from_numpy(self.searchspace.comp_rep_bounds.values)

    def get_numerical_indices(self, n_inputs: int) -> tuple[int, ...]:
        """Get the indices of the regular numerical model inputs."""
        return tuple(i for i in range(n_inputs) if i != self.task_idx)


@define
class GaussianProcessSurrogate(Surrogate):
    """A Gaussian process surrogate model."""

    # Note [Scaling Workaround]
    # -------------------------
    # For GPs, we deactivate the base class scaling and instead let the botorch
    # model internally handle input/output scaling. The reasons is that we need to
    # make `to_botorch` expose the actual botorch GP object, instead of going
    # via the `AdapterModel`, because certain acquisition functions (like qNIPV)
    # require the capability to `fantasize`, which the `AdapterModel` does not support.
    # The base class scaling thus needs to be disabled since otherwise the botorch GP
    # object would be trained on pre-scaled input/output data. This would cause a
    # problem since the resulting `posterior` method of that object is exposed
    # to `optimize_acqf_*`, which is configured to be called on the original scale.
    # Moving the scaling operation into the botorch GP object avoids this conflict.

    supports_transfer_learning: ClassVar[bool] = True
    # See base class.

    kernel_factory: KernelFactory = field(
        alias="kernel_or_factory",
        factory=DefaultKernelFactory,
        converter=to_kernel_factory,
    )
    """The factory used to create the kernel of the Gaussian process.

    Accepts either a :class:`baybe.kernels.base.Kernel` or a
    :class:`.kernel_factory.KernelFactory`.
    When passing a :class:`baybe.kernels.base.Kernel`, it gets automatically wrapped
    into a :class:`.kernel_factory.PlainKernelFactory`."""

    # TODO: type should be Optional[botorch.models.SingleTaskGP] but is currently
    #   omitted due to: https://github.com/python-attrs/cattrs/issues/531
    _model = field(init=False, default=None, eq=False)
    """The actual model."""

    @staticmethod
    def from_preset(preset: GaussianProcessPreset) -> GaussianProcessSurrogate:
        """Create a Gaussian process surrogate from one of the defined presets."""
        return make_gp_from_preset(preset)

    def to_botorch(self) -> Model:  # noqa: D102
        # See base class.

        return self._model

    @staticmethod
    def _make_parameter_scaler_factory(
        parameter: Parameter,
    ) -> type[InputTransform] | None:
        # See base class.

        # For GPs, we let botorch handle the scaling. See [Scaling Workaround] above.
        return None

    @staticmethod
    def _make_target_scaler_factory() -> type[OutcomeTransform] | None:
        # See base class.

        # For GPs, we let botorch handle the scaling. See [Scaling Workaround] above.
        return None

    def _posterior(self, candidates_comp_scaled: Tensor, /) -> Posterior:
        # See base class.
        return self._model.posterior(candidates_comp_scaled)

    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        # See base class.

        import botorch
        import gpytorch
        import torch

        # FIXME[typing]: It seems there is currently no better way to inform the type
        #   checker that the attribute is available at the time of the function call
        assert self._searchspace is not None

        context = _ModelContext(self._searchspace)

        numerical_idxs = context.get_numerical_indices(train_x.shape[-1])

        # For GPs, we let botorch handle the scaling. See [Scaling Workaround] above.
        input_transform = botorch.models.transforms.Normalize(
            train_x.shape[-1], bounds=context.parameter_bounds, indices=numerical_idxs
        )
        outcome_transform = botorch.models.transforms.Standardize(train_y.shape[-1])

        # extract the batch shape of the training data
        batch_shape = train_x.shape[:-2]

        # create GP mean
        mean_module = gpytorch.means.ConstantMean(batch_shape=batch_shape)

        # define the covariance module for the numeric dimensions
        base_covar_module = self.kernel_factory(
            context.searchspace, train_x, train_y
        ).to_gpytorch(
            ard_num_dims=train_x.shape[-1] - context.n_task_dimensions,
            active_dims=numerical_idxs,
            batch_shape=batch_shape,
        )

        # create GP covariance
        if not context.is_multitask:
            covar_module = base_covar_module
        else:
            task_covar_module = gpytorch.kernels.IndexKernel(
                num_tasks=context.n_tasks,
                active_dims=context.task_idx,
                rank=context.n_tasks,  # TODO: make controllable
            )
            covar_module = base_covar_module * task_covar_module

        # create GP likelihood
        noise_prior = _default_noise_factory(context.searchspace, train_x, train_y)
        likelihood = gpytorch.likelihoods.GaussianLikelihood(
            noise_prior=noise_prior[0].to_gpytorch(), batch_shape=batch_shape
        )
        likelihood.noise = torch.tensor([noise_prior[1]])

        # construct and fit the Gaussian process
        self._model = botorch.models.SingleTaskGP(
            train_x,
            train_y,
            input_transform=input_transform,
            outcome_transform=outcome_transform,
            mean_module=mean_module,
            covar_module=covar_module,
            likelihood=likelihood,
        )
        mll = gpytorch.ExactMarginalLogLikelihood(self._model.likelihood, self._model)

        # TODO: This is a simple temporary workaround to avoid model overfitting
        #   via early stopping in the presence of task parameters, which currently
        #   have no prior configured.
        if context.n_task_dimensions > 0:
            botorch.optim.fit.fit_gpytorch_mll_torch(mll, step_limit=200)
        else:
            botorch.fit.fit_gpytorch_mll(mll)

    def __str__(self) -> str:
        fields = [
            to_string("Kernel factory", self.kernel_factory, single_line=True),
        ]
        return to_string(super().__str__(), *fields)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Validation tests for composite kernels."""

import pytest
from pytest import param

from baybe.kernels.basic import MaternKernel
from baybe.kernels.composite import AdditiveKernel, ProductKernel, ScaleKernel

base_kernel = MaternKernel()


@pytest.mark.parametrize(
    ("base_kernel", "outputscale_prior", "outputscale_initial_value", "error"),
    [
        param("wrong", None, None, TypeError, id="not_a_kernel"),
        param(base_kernel, "wrong", None, TypeError, id="not_a_prior"),
        param(base_kernel, None, "wrong", ValueError, id="not_a_scale"),
        param(base_kernel, None, -1, ValueError, id="negative_scale"),
        param(base_kernel, None, 0, ValueError, id="zero_scale"),
    ],
)
def test_scale_kernel_validation(
    base_kernel, outputscale_prior, outputscale_initial_value, error
):
    """Providing invalid kernel arguments raises an exception."""
    with pytest.raises(error):
        ScaleKernel(base_kernel, outputscale_prior, outputscale_initial_value)


@pytest.mark.parametrize("kernel_cls", [AdditiveKernel, ProductKernel])
@pytest.mark.parametrize(
    ("base_kernels", "error"),
    [
        param(1, TypeError, id="not_an_iterable"),
        param([], ValueError, id="empty_iterable"),
        param([base_kernel], ValueError, id="only_one_kernel"),
        param([1, base_kernel], TypeError, id="not_a_kernel"),
    ],
)
def test_additive_and_product_kernel_validation(kernel_cls, base_kernels, error):
    """Providing invalid kernel arguments raises an exception."""
    with pytest.raises(error):
        kernel_cls(base_kernels)



"""Optional chemistry imports."""

from baybe.exceptions import OptionalImportError

try:
    from mordred import Calculator, descriptors
    from rdkit import Chem, RDLogger
    from rdkit.Chem.rdMolDescriptors import GetMorganFingerprintAsBitVect
except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Chemistry functionality is unavailable because the necessary optional "
        "dependencies are not installed. "
        "Consider installing BayBE with 'chem' dependency, "
        "e.g. via `pip install baybe[chem]`."
    ) from ex

__all__ = [
    "descriptors",
    "Calculator",
    "Chem",
    "GetMorganFingerprintAsBitVect",
    "RDLogger",
]



"""EDBO preset for Gaussian process surrogates."""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING

from attrs import define

from baybe.kernels.basic import MaternKernel
from baybe.kernels.composite import ScaleKernel
from baybe.parameters import TaskParameter
from baybe.priors.basic import GammaPrior
from baybe.surrogates.gaussian_process.kernel_factory import KernelFactory

if TYPE_CHECKING:
    from torch import Tensor

    from baybe.kernels.base import Kernel
    from baybe.searchspace.core import SearchSpace


@define
class EDBOKernelFactory(KernelFactory):
    """A factory providing the kernel for Gaussian process surrogates adapted from EDBO.

    References:
        * https://github.com/b-shields/edbo/blob/master/edbo/bro.py#L664
        * https://doi.org/10.1038/s41586-021-03213-y
    """

    def __call__(  # noqa: D102
        self, searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
    ) -> Kernel:
        # See base class.
        effective_dims = train_x.shape[-1] - len(
            [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
        )

        mordred = (searchspace.contains_mordred or searchspace.contains_rdkit) and (
            effective_dims >= 50
        )

        # low D priors
        if effective_dims < 5:
            lengthscale_prior = GammaPrior(1.2, 1.1)
            lengthscale_initial_value = 0.2
            outputscale_prior = GammaPrior(5.0, 0.5)
            outputscale_initial_value = 8.0

        # DFT optimized priors
        elif mordred and effective_dims < 100:
            lengthscale_prior = GammaPrior(2.0, 0.2)
            lengthscale_initial_value = 5.0
            outputscale_prior = GammaPrior(5.0, 0.5)
            outputscale_initial_value = 8.0

        # Mordred optimized priors
        elif mordred:
            lengthscale_prior = GammaPrior(2.0, 0.1)
            lengthscale_initial_value = 10.0
            outputscale_prior = GammaPrior(2.0, 0.1)
            outputscale_initial_value = 10.0

        # OHE optimized priors
        else:
            lengthscale_prior = GammaPrior(3.0, 1.0)
            lengthscale_initial_value = 2.0
            outputscale_prior = GammaPrior(5.0, 0.2)
            outputscale_initial_value = 20.0

        return ScaleKernel(
            MaternKernel(
                nu=2.5,
                lengthscale_prior=lengthscale_prior,
                lengthscale_initial_value=lengthscale_initial_value,
            ),
            outputscale_prior=outputscale_prior,
            outputscale_initial_value=outputscale_initial_value,
        )


def _edbo_noise_factory(
    searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
) -> tuple[GammaPrior, float]:
    """Create the default noise settings for the Gaussian process surrogate.

    The logic is adapted from EDBO (Experimental Design via Bayesian Optimization).

    References:
        * https://github.com/b-shields/edbo/blob/master/edbo/bro.py#L664
        * https://doi.org/10.1038/s41586-021-03213-y
    """
    # TODO: Replace this function with a proper likelihood factory
    effective_dims = train_x.shape[-1] - len(
        [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
    )

    uses_descriptors = (
        searchspace.contains_mordred or searchspace.contains_rdkit
    ) and (effective_dims >= 50)

    # low D priors
    if effective_dims < 5:
        return (GammaPrior(1.05, 0.5), 0.1)

    # DFT optimized priors
    elif uses_descriptors and effective_dims < 100:
        return (GammaPrior(1.5, 0.1), 5.0)

    # Mordred optimized priors
    elif uses_descriptors:
        return (GammaPrior(1.5, 0.1), 5.0)

    # OHE optimized priors
    else:
        return (GammaPrior(1.5, 0.1), 5.0)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Test serialization of kernels."""

from hypothesis import given

from baybe.kernels.base import Kernel
from tests.hypothesis_strategies.kernels import kernels


@given(kernels())
def test_kernel_roundtrip(kernel: Kernel):
    string = kernel.to_json()
    kernel2 = Kernel.from_json(string)
    assert kernel == kernel2, (kernel, kernel2)



"""BayBE objectives."""

from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective

__all__ = [
    "SingleTargetObjective",
    "DesirabilityObjective",
]



"""Binary targets."""

import gc
from typing import TypeAlias

import numpy as np
import pandas as pd
from attrs import define, field
from attrs.validators import instance_of

from baybe.exceptions import InvalidTargetValueError
from baybe.serialization import SerialMixin
from baybe.targets.base import Target
from baybe.utils.validation import validate_not_nan

ChoiceValue: TypeAlias = bool | int | float | str
"""Types of values that a :class:`BinaryTarget` can take."""

_SUCCESS_VALUE_COMP = 1.0
"""Computational representation of the success value."""

_FAILURE_VALUE_COMP = 0.0
"""Computational representation of the failure value."""


@define(frozen=True)
class BinaryTarget(Target, SerialMixin):
    """Class for binary targets."""

    # FIXME[typing]: https://github.com/python-attrs/attrs/issues/1336

    success_value: ChoiceValue = field(
        default=True,
        validator=[instance_of(ChoiceValue), validate_not_nan],  # type: ignore[call-overload]
        kw_only=True,
    )
    """Experimental representation of the success value."""

    failure_value: ChoiceValue = field(
        default=False,
        validator=[instance_of(ChoiceValue), validate_not_nan],  # type: ignore[call-overload]
        kw_only=True,
    )
    """Experimental representation of the failure value."""

    @failure_value.validator
    def _validate_values(self, _, value):
        """Validate that the two choice values of the target are different."""
        if value == self.success_value:
            raise ValueError(
                f"The two choice values of a '{BinaryTarget.__name__}' must be "
                f"different but the following value was provided for both choices of "
                f"target '{self.name}': {value}"
            )

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:  # noqa: D102
        # TODO: The method (signature) needs to be refactored, potentially when
        #   enabling multi-target settings. The current input type suggests that passing
        #   dataframes is allowed, but the code was designed for single targets and
        #   desirability objectives, where only one column is present.
        assert data.shape[1] == 1

        # Validate target values
        col = data.iloc[:, [0]]
        invalid = col[~col.isin([self.success_value, self.failure_value]).values]
        if len(invalid) > 0:
            raise InvalidTargetValueError(
                f"The following values entered for target '{self.name}' are not in the "
                f"set of accepted choice values "
                f"{{self.success_value, self.failure_value}}: \n{invalid}"
            )

        # Transform
        success_idx = data.iloc[:, 0] == self.success_value
        return pd.DataFrame(
            np.where(success_idx, _SUCCESS_VALUE_COMP, _FAILURE_VALUE_COMP),
            index=data.index,
            columns=data.columns,
        )

    def summary(self) -> dict:  # noqa: D102
        # See base class.
        return dict(
            Type=self.__class__.__name__,
            Name=self.name,
            Success_value=self.success_value,
            Failure_value=self.failure_value,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Acquisition function wrappers."""

from baybe.acquisition.acqfs import (
    ExpectedImprovement,
    LogExpectedImprovement,
    PosteriorMean,
    PosteriorStandardDeviation,
    ProbabilityOfImprovement,
    UpperConfidenceBound,
    qExpectedImprovement,
    qKnowledgeGradient,
    qLogExpectedImprovement,
    qLogNoisyExpectedImprovement,
    qNegIntegratedPosteriorVariance,
    qNoisyExpectedImprovement,
    qProbabilityOfImprovement,
    qSimpleRegret,
    qThompsonSampling,
    qUpperConfidenceBound,
)

PM = PosteriorMean
PSTD = PosteriorStandardDeviation
qSR = qSimpleRegret
EI = ExpectedImprovement
qEI = qExpectedImprovement
qKG = qKnowledgeGradient
LogEI = LogExpectedImprovement
qLogEI = qLogExpectedImprovement
qNEI = qNoisyExpectedImprovement
qNIPV = qNegIntegratedPosteriorVariance
qLogNEI = qLogNoisyExpectedImprovement
PI = ProbabilityOfImprovement
qPI = qProbabilityOfImprovement
UCB = UpperConfidenceBound
qUCB = qUpperConfidenceBound
qTS = qThompsonSampling

__all__ = [
    ######################### Acquisition functions
    # Knowledge Gradient
    "qKnowledgeGradient",
    # Posterior Statistics
    "PosteriorMean",
    "PosteriorStandardDeviation",
    # Simple Regret
    "qSimpleRegret",
    # Expected Improvement
    "ExpectedImprovement",
    "qExpectedImprovement",
    "LogExpectedImprovement",
    "qLogExpectedImprovement",
    "qNoisyExpectedImprovement",
    "qNegIntegratedPosteriorVariance",
    "qLogNoisyExpectedImprovement",
    # Probability of Improvement
    "ProbabilityOfImprovement",
    "qProbabilityOfImprovement",
    # Upper Confidence Bound
    "UpperConfidenceBound",
    "qUpperConfidenceBound",
    # Thompson Sampling
    "qThompsonSampling",
    ######################### Abbreviations
    # Knowledge Gradient
    "qKG",
    # Posterior Statistics
    "PM",
    "PSTD",
    # Simple Regret
    "qSR",
    # Expected Improvement
    "EI",
    "qEI",
    "LogEI",
    "qLogEI",
    "qNEI",
    "qNIPV",
    "qLogNEI",
    # Probability of Improvement
    "PI",
    "qPI",
    # Upper Confidence Bound
    "UCB",
    "qUCB",
    # Thompson Sampling
    "qTS",
]



"""Validation functionality for search spaces."""

from collections.abc import Collection, Sequence
from typing import TypeVar

import pandas as pd

from baybe.exceptions import EmptySearchSpaceError
from baybe.parameters import TaskParameter
from baybe.parameters.base import Parameter

_T = TypeVar("_T", bound=Parameter)


def validate_parameter_names(  # noqa: DOC101, DOC103
    parameters: Collection[Parameter],
) -> None:
    """Validate the parameter names.

    Raises:
        ValueError: If the given list contains parameters with the same name.
    """
    param_names = [p.name for p in parameters]
    if len(set(param_names)) != len(param_names):
        raise ValueError("All parameters must have unique names.")


def validate_parameters(parameters: Collection[Parameter]) -> None:  # noqa: DOC101, DOC103
    """Validate the parameters.

    Raises:
        EmptySearchSpaceError: If the parameter list is empty.
        NotImplementedError: If more than one
            :class:`baybe.parameters.categorical.TaskParameter` is requested.
    """
    if not parameters:
        raise EmptySearchSpaceError("At least one parameter must be provided.")

    # TODO [16932]: Remove once more task parameters are supported
    if len([p for p in parameters if isinstance(p, TaskParameter)]) > 1:
        raise NotImplementedError(
            "Currently, at most one task parameter can be considered."
        )

    # Assert: unique names
    validate_parameter_names(parameters)


def get_transform_parameters(
    parameters: Sequence[_T],
    df: pd.DataFrame,
    allow_missing: bool,
    allow_extra: bool,
) -> list[_T]:
    """Extract the parameters relevant for transforming a given dataframe.

    Args:
        parameters: The parameters to be considered for transformation (provided
            they have match in the given dataframe).
        df: See :meth:`baybe.searchspace.core.SearchSpace.transform`.
        allow_missing: See :meth:`baybe.searchspace.core.SearchSpace.transform`.
        allow_extra: See :meth:`baybe.searchspace.core.SearchSpace.transform`.

    Raises:
        ValueError: If the given parameters and dataframe are not compatible
            under the specified values for the Boolean flags.

    Returns:
        The (subset of) parameters that need to be considered for the transformation.
    """
    parameter_names = [p.name for p in parameters]

    if (not allow_missing) and (missing := set(parameter_names) - set(df)):  # type: ignore[arg-type]
        raise ValueError(
            f"The search space parameter(s) {missing} cannot be matched against "
            f"the provided dataframe. If you want to transform a subset of "
            f"parameter columns, explicitly set `allow_missing=True`."
        )

    if (not allow_extra) and (extra := set(df) - set(parameter_names)):
        raise ValueError(
            f"The provided dataframe column(s) {extra} cannot be matched against"
            f"the search space parameters. If you want to transform a dataframe "
            f"with additional columns, explicitly set `allow_extra=True'."
        )

    return (
        [p for p in parameters if p.name in df] if allow_missing else list(parameters)
    )



"""Integration tests."""

import numpy as np
import pandas as pd
import pytest
from pytest import param

from baybe.parameters.numerical import NumericalDiscreteParameter
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace.core import SearchSpace
from baybe.targets.numerical import NumericalTarget
from baybe.utils.basic import get_subclasses

nonpredictive_recommenders = [
    param(cls(allow_recommending_already_measured=True), id=cls.__name__)
    for cls in get_subclasses(NonPredictiveRecommender)
]

p1 = NumericalDiscreteParameter("p1", [1, 2])
t1 = NumericalTarget("t1", "MAX")
objective = t1.to_objective()
measurements = pd.DataFrame(
    {p1.name: p1.values, t1.name: np.random.random(len(p1.values))}
)


@pytest.fixture(name="searchspace")
def fixture_searchspace():
    return SearchSpace.from_product([p1])


@pytest.mark.parametrize("recommender", nonpredictive_recommenders)
def test_nonbayesian_recommender_with_measurements(recommender, searchspace):
    """Calling a non-Bayesian recommender with training data raises a warning."""
    with pytest.warns(
        UserWarning,
        match=(
            f"'{recommender.__class__.__name__}' does not utilize any training data"
        ),
    ):
        recommender.recommend(
            batch_size=1, searchspace=searchspace, measurements=measurements
        )


@pytest.mark.parametrize("recommender", nonpredictive_recommenders)
def test_nonbayesian_recommender_with_objective(recommender, searchspace):
    """Calling a non-Bayesian recommender with an objective raises a warning."""
    with pytest.warns(
        UserWarning,
        match=(f"'{recommender.__class__.__name__}' does not consider any objectives"),
    ):
        recommender.recommend(
            batch_size=1, searchspace=searchspace, objective=objective
        )



"""Optional xyzpy import."""

from baybe.exceptions import OptionalImportError

try:
    import xyzpy
except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Batch scenario simulation is unavailable because 'xyzpy' is not "
        "installed. "
        "Consider installing BayBE with 'simulation' dependency, "
        "e.g. via `pip install baybe[simulation]`."
    ) from ex

__all__ = [
    "xyzpy",
]



"""This file contains some utility function for building the documentation."""

import re
import sys
from pathlib import Path

from packaging.version import Version


def adjust_pictures(
    file_path: str, match: str, light_version: str, dark_version: str
) -> None:
    """Adjust pictures to have different versions for light and dark furo style.

    Args:
        file_path: The (relative) path to the file that needs to be adjusted. Typically
            the index and README_link file.
        match: The expression that identifies the picture name as referenced in the
            file (do not include the file extension)
        light_version: The name of the light mode picture version.
        dark_version: The name of the dark mode picture version.
    """
    with open(file_path) as file:
        lines = file.readlines()

    line_index = None
    for i, line in enumerate(lines):
        if match in line:
            line_index = i
            break

    if line_index is not None:
        line = lines[line_index]
        light_line = line
        light_line = light_line.replace(  # For replacing the banner
            '"reference external"', '"reference external only-light"'
        )
        light_line = light_line.replace(  # For replacing the example plot
            'img alt="Substance Encoding Example" ',
            'img alt="Substance Encoding Example" class="only-light align-center" ',
        )
        lines[line_index] = light_line
        dark_line = light_line.replace("light", "dark")
        dark_line = dark_line.replace(light_version, dark_version)
        lines.insert(line_index + 1, dark_line)

        with open(file_path, "w") as file:
            file.writelines(lines)


def add_version_to_selector_page(version: str) -> None:
    """Add the newly built version to the version selection overview.

    Args:
        version: The version that should be added.
    """
    indent = "        "
    new_line = (
        f"{indent}<li><a href="
        f'"https://emdgroup.github.io/baybe/{version}/">{version}'
        "</a></li>\n"
    )
    file = Path("versions.html")
    modified_lines = []

    # Regular expression pattern to match the version number in the format X.Y.Z
    pattern = r"\b\d+\.\d+\.\d+\b"

    # Indices for memorizing the index of the line containing "stable" and the last
    # version that was larger than the one to be inserted
    stable_index = -1
    last_version_index = -1
    with file.open(mode="r") as f:
        lines = f.readlines()
        for ind, line in enumerate(lines):
            # We might find strings that look like version numbers without wanting to
            # We thus need to potentially reset `last_version_index`
            if "<h1>Available versions</h1>" in line:
                last_version_index = -1

            # Search for the version number in the current line
            if version_number := re.search(pattern, line):
                # Memorize whether we saw a larger version number, implying that this is
                # a hotfix
                if Version(version_number.group()) > Version(version):
                    last_version_index = ind

            # Add the already existing line
            modified_lines.append(line)

            # Slightly weird string since the word "stable" appears is other places
            if "stable</a></li>" in line:
                stable_index = ind

    # We never found a larger number than the current, so we just insert after stable
    index = last_version_index + 1 if last_version_index != -1 else stable_index + 1
    modified_lines.insert(index, new_line)
    with file.open(mode="w") as f:
        f.writelines(modified_lines)


def adjust_version_to_html(version: str) -> None:
    """Adjust the shown version in the HTML files.

    This method includes the current version in the sidebar of all HTML files and
    additionally adds a banner warning when the documentation is not ``stable``.

    Args:
        version: The version that should be injected into the sidebar.
    """
    prefix = '<li class="toctree-l1">'
    html_class = "reference external"
    link = "https://emdgroup.github.io/baybe/versions"
    new_line = (
        f'{prefix}<a class={html_class} href="{link}">Version: {version}</a></li>'
    )
    link_to_stable = '<a href="https://emdgroup.github.io/baybe/stable">stable</a>'
    # The content of the announcement that might be added.
    announcement = (
        f"You are not viewing the documentation of the {link_to_stable} version."
    )
    # Actual HTML code that will be inserted
    announcement_html = (
        """<div class="announcement">\n"""
        """    <aside class="announcement-content">\n"""
        f"""        {announcement}\n"""
        """    </aside>\n"""
        """</div>\n"""
    )
    add_announcement = version != "stable"
    path = Path(version)
    if path.exists():
        # Recursively check all HTML files
        for file in path.rglob("*.html"):
            modified_lines = []
            with file.open(mode="r") as f:
                lines = f.readlines()
                for line in lines:
                    # Check if we need to add the announcement
                    if add_announcement:
                        # Add announcement at correct position
                        if line.strip() == '<div class="page">':
                            modified_lines.insert(-2, announcement_html)
                    if "Versions</a></li>" in line:
                        modified_lines.append(new_line)
                    else:
                        modified_lines.append(line)
            with file.open(mode="w") as f:
                f.writelines(modified_lines)


def check_for_hotfix(tags: list[str], version: str):
    """Check whether the current build corresponds to a hotfix."""
    split_tags = tags.split("\n")
    split_tags.sort(key=Version)
    print(Version(version) < Version(split_tags[-1]), end="")


if __name__ == "__main__":
    chosen_method = sys.argv[1]
    version = sys.argv[2]
    if chosen_method == "selector_page":
        add_version_to_selector_page(version)
    elif chosen_method == "html":
        adjust_version_to_html(version)
    elif chosen_method == "hotfix":
        check_for_hotfix(tags=sys.argv[3], version=version)
    else:
        print(f"Invalid arguments {sys.argv} were chosen!")



"""Objective-related enumerations."""

from enum import Enum


class Scalarizer(Enum):
    """Available scalarization mechanisms for desirability objectives."""

    MEAN = "MEAN"
    """Arithmetic mean."""

    GEOM_MEAN = "GEOM_MEAN"
    """Geometric mean."""



"""Functionality for single-target objectives."""

import gc

import pandas as pd
from attr import define, field
from attr.validators import instance_of

from baybe.objectives.base import Objective
from baybe.targets.base import Target
from baybe.utils.dataframe import pretty_print_df
from baybe.utils.plotting import to_string


@define(frozen=True, slots=False)
class SingleTargetObjective(Objective):
    """An objective focusing on a single target."""

    _target: Target = field(validator=instance_of(Target), alias="target")
    """The single target considered by the objective."""

    def __str__(self) -> str:
        targets_list = [target.summary() for target in self.targets]
        targets_df = pd.DataFrame(targets_list)

        fields = [
            to_string("Type", self.__class__.__name__, single_line=True),
            to_string("Targets", pretty_print_df(targets_df)),
        ]

        return to_string("Objective", *fields)

    @property
    def targets(self) -> tuple[Target, ...]:  # noqa: D102
        # See base class.
        return (self._target,)

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:  # noqa: D102
        # See base class.
        target_data = data[[self._target.name]].copy()
        return self._target.transform(target_data)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Base class for all nonpredictive recommenders."""

import gc
import warnings
from abc import ABC

import pandas as pd
from attr import fields
from attrs import define

from baybe.exceptions import UnusedObjectWarning
from baybe.objectives.base import Objective
from baybe.recommenders.pure.base import PureRecommender
from baybe.searchspace.core import SearchSpace, SearchSpaceType


@define
class NonPredictiveRecommender(PureRecommender, ABC):
    """Abstract base class for all nonpredictive recommenders."""

    def recommend(  # noqa: D102
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        # See base class.

        if (measurements is not None) and (len(measurements) != 0):
            warnings.warn(
                f"'{self.recommend.__name__}' was called with a non-empty "
                f"set of measurements but '{self.__class__.__name__}' does not "
                f"utilize any training data, meaning that the argument is ignored.",
                UnusedObjectWarning,
            )
        if objective is not None:
            warnings.warn(
                f"'{self.recommend.__name__}' was called with a an explicit objective "
                f"but '{self.__class__.__name__}' does not "
                f"consider any objectives, meaning that the argument is ignored.",
                UnusedObjectWarning,
            )
        if (pending_experiments is not None) and (
            self.allow_recommending_pending_experiments
            or searchspace.type is not SearchSpaceType.DISCRETE
        ):
            warnings.warn(
                f"Pending experiments were provided but the selected recommender "
                f"'{self.__class__.__name__}' only utilizes this information for "
                f"purely discrete spaces and "
                f"{fields(self.__class__).allow_recommending_pending_experiments.name}"
                f"=False.",
                UnusedObjectWarning,
            )
        return super().recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=pending_experiments,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Conversion utilities."""

from collections.abc import Sequence
from fractions import Fraction
from typing import TypeVar

from attrs import Attribute

_T = TypeVar("_T")


def fraction_to_float(value: str | float | Fraction, /) -> float:
    """Convert the provided input representing a fraction into a float.

    Args:
        value: The input to be converted.

    Returns:
        The float representation of the given input.

    Raises:
        ValueError: If the input was provided as string but could not be interpreted as
            fraction.
    """
    if isinstance(value, str):
        try:
            value = Fraction(value)
        except ValueError as err:
            raise ValueError(
                f"The provided input '{value}' could not be interpreted as a fraction."
            ) from err
    return float(value)


def nonstring_to_tuple(x: Sequence[_T], self: type, field: Attribute) -> tuple[_T, ...]:
    """Convert a sequence to tuple but raise an exception for string input."""
    if isinstance(x, str):
        raise ValueError(
            f"Argument passed to '{field.alias}' of class '{self.__class__.__name__}' "
            f"must be a sequence but cannot be a string."
        )
    return tuple(x)



"""Tests for the objective module."""

import numpy as np
import pytest
from cattrs import IterableValidationError

from baybe.objectives.desirability import DesirabilityObjective, scalarize
from baybe.objectives.enum import Scalarizer
from baybe.objectives.single import SingleTargetObjective
from baybe.targets import NumericalTarget


class TestInvalidObjectiveCreation:
    """Invalid objective creation raises expected error."""

    # Two example targets used in the tests
    two_targets = [
        NumericalTarget(
            name="Target_1",
            mode="MAX",
            bounds=(0, 100),
        ),
        NumericalTarget(
            name="Target_2",
            mode="MIN",
            bounds=(0, 100),
        ),
    ]

    def test_empty_target_list(self):
        with pytest.raises(ValueError):
            DesirabilityObjective(targets=[])

    def test_wrong_target_type(self):
        with pytest.raises(TypeError):
            SingleTargetObjective(target={"A": 1, "B": 2})

    def test_missing_bounds_for_desirability(self):
        with pytest.raises(ValueError):
            DesirabilityObjective(
                targets=[
                    NumericalTarget(
                        name="Target_1",
                        mode="MAX",
                        bounds=(0, 100),
                    ),
                    NumericalTarget(
                        name="Target_2",
                        mode="MIN",
                    ),
                ],
            )

    def test_invalid_combination_function(self):
        with pytest.raises(ValueError):
            DesirabilityObjective(
                targets=self.two_targets,
                scalarizer="FALSE_STUFF",
            )

    def test_wrong_number_of_weights(self):
        with pytest.raises(ValueError):
            DesirabilityObjective(
                targets=self.two_targets,
                weights=[1, 2, 3],
            )

    def test_non_numeric_weights(self):
        with pytest.raises(IterableValidationError):
            DesirabilityObjective(
                targets=self.two_targets,
                weights=[1, "ABC"],
            )

    def test_wrong_weights_type(self):
        with pytest.raises(IterableValidationError):
            DesirabilityObjective(
                targets=self.two_targets,
                weights="ABC",
            )


@pytest.mark.parametrize(
    ("values", "scalarizer", "weights", "expected"),
    [
        ([[1, 2]], Scalarizer.MEAN, [1, 1], [1.5]),
        ([[1, 2]], Scalarizer.MEAN, [1, 2], [5 / 3]),
        ([[1, 2]], Scalarizer.GEOM_MEAN, [1, 1], [np.sqrt(2)]),
        ([[1, 2]], Scalarizer.GEOM_MEAN, [1, 2], [np.power(4, 1 / 3)]),
    ],
)
def test_desirability_scalarization(values, scalarizer, weights, expected):
    """The desirability scalarization yields the expected result."""
    actual = scalarize(values, scalarizer, weights)
    assert np.array_equal(actual, expected), (expected, actual)



"""Temporary dispatching functionality for backward compatibility."""

import warnings
from typing import Any, Literal

from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.targets.base import Target


def Objective(
    mode: Literal["SINGLE", "DESIRABILITY"],
    targets: list[Target],
    weights: list[float] | None = None,
    combine_func: Literal["MEAN", "GEOM_MEAN"] | None = None,
):
    """Return the appropriate new-style class depending on the mode."""
    warnings.warn(
        "The use of `baybe.objective` is deprecated and will be disabled in "
        "a future version. Use the classes defined in `baybe.objectives` instead.",
        DeprecationWarning,
    )

    if mode == "SINGLE":
        if len(targets) != 1:
            raise ValueError("In 'SINGLE' mode, you must provide exactly one target.")
        return SingleTargetObjective(targets[0])

    elif mode == "DESIRABILITY":
        kwargs: dict[str, Any] = {}
        if weights is not None:
            kwargs["weights"] = weights
        if combine_func is not None:
            kwargs["scalarizer"] = combine_func
        return DesirabilityObjective(targets, **kwargs)

    else:
        raise ValueError(f"Unknown mode: {mode}")



"""Base classes for all constraints."""

from __future__ import annotations

import gc
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar

import pandas as pd
from attr import define, field
from attr.validators import ge, instance_of, min_len

from baybe.constraints.deprecation import structure_constraints
from baybe.serialization import (
    SerialMixin,
    converter,
    unstructure_base,
)

if TYPE_CHECKING:
    import polars as pl


@define
class Constraint(ABC, SerialMixin):
    """Abstract base class for all constraints."""

    # class variables
    # TODO: it might turn out these are not needed at a later development stage
    eval_during_creation: ClassVar[bool]
    """Class variable encoding whether the condition is evaluated during creation."""

    eval_during_modeling: ClassVar[bool]
    """Class variable encoding whether the condition is evaluated during modeling."""

    eval_during_augmentation: ClassVar[bool] = False
    """Class variable encoding whether the constraint could be considered during data
    augmentation."""

    numerical_only: ClassVar[bool] = False
    """Class variable encoding whether the constraint is valid only for numerical
    parameters."""

    # Object variables
    parameters: list[str] = field(validator=min_len(1))
    """The list of parameters used for the constraint."""

    @parameters.validator
    def _validate_params(  # noqa: DOC101, DOC103
        self, _: Any, params: list[str]
    ) -> None:
        """Validate the parameter list.

        Raises:
            ValueError: If ``params`` contains duplicate values.
        """
        if len(params) != len(set(params)):
            raise ValueError(
                f"The given 'parameters' list must have unique values "
                f"but was: {params}."
            )

    def summary(self) -> dict:
        """Return a custom summarization of the constraint."""
        constr_dict = dict(
            Type=self.__class__.__name__, Affected_Parameters=self.parameters
        )
        return constr_dict

    @property
    def is_continuous(self) -> bool:
        """Boolean indicating if this is a constraint over continuous parameters."""
        return isinstance(self, ContinuousConstraint)

    @property
    def is_discrete(self) -> bool:
        """Boolean indicating if this is a constraint over discrete parameters."""
        return isinstance(self, DiscreteConstraint)


@define
class DiscreteConstraint(Constraint, ABC):
    """Abstract base class for discrete constraints.

    Discrete constraints use conditions and chain them together to filter unwanted
    entries from the search space.
    """

    # class variables
    eval_during_creation: ClassVar[bool] = True
    # See base class.

    eval_during_modeling: ClassVar[bool] = False
    # See base class.

    @abstractmethod
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        """Get the indices of dataframe entries that are invalid under the constraint.

        Args:
            data: A dataframe where each row represents a particular parameter
                combination.

        Returns:
            The dataframe indices of rows where the constraint is violated.
        """

    def get_invalid_polars(self) -> pl.Expr:
        """Translate the constraint to Polars expression identifying undesired rows.

        Returns:
            The Polars expressions to pass to :meth:`polars.LazyFrame.filter`.

        Raises:
            NotImplementedError: If the constraint class does not have a Polars
                implementation.
        """
        raise NotImplementedError(
            f"'{self.__class__.__name__}' does not have a Polars implementation."
        )


@define
class ContinuousConstraint(Constraint, ABC):
    """Abstract base class for continuous constraints."""

    # class variables
    eval_during_creation: ClassVar[bool] = False
    # See base class.

    eval_during_modeling: ClassVar[bool] = True
    # See base class.

    numerical_only: ClassVar[bool] = True
    # See base class.


@define
class CardinalityConstraint(Constraint, ABC):
    """Abstract base class for cardinality constraints.

    Places a constraint on the set of nonzero (i.e. "active") values among the
    specified parameters, bounding it between the two given integers,
        ``min_cardinality`` <= |{p_i : p_i != 0}| <= ``max_cardinality``
    where ``{p_i}`` are the parameters specified for the constraint.

    Note that this can be equivalently regarded as L0-constraint on the vector
    containing the specified parameters.
    """

    # class variable
    numerical_only: ClassVar[bool] = True
    # See base class.

    # object variables
    min_cardinality: int = field(default=0, validator=[instance_of(int), ge(0)])
    "The minimum required cardinality."

    max_cardinality: int = field(validator=instance_of(int))
    "The maximum allowed cardinality."

    @max_cardinality.default
    def _default_max_cardinality(self):
        """Use the number of involved parameters as the upper limit by default."""
        return len(self.parameters)

    def __attrs_post_init__(self):
        """Validate the cardinality bounds.

        Raises:
            ValueError: If the provided cardinality bounds are invalid.
            ValueError: If the provided cardinality bounds impose no constraint.
        """
        if self.min_cardinality > self.max_cardinality:
            raise ValueError(
                f"The lower cardinality bound cannot be larger than the upper bound. "
                f"Provided values: {self.max_cardinality=}, {self.min_cardinality=}."
            )

        if self.max_cardinality > len(self.parameters):
            raise ValueError(
                f"The cardinality bound cannot exceed the number of parameters. "
                f"Provided values: {self.max_cardinality=}, {len(self.parameters)=}."
            )

        if self.min_cardinality == 0 and self.max_cardinality == len(self.parameters):
            raise ValueError(
                f"No constraint of type `{self.__class__.__name__}' is required "
                f"when the lower cardinality bound is zero and the upper bound equals "
                f"the number of parameters. Provided values: {self.min_cardinality=}, "
                f"{self.max_cardinality=}, {len(self.parameters)=}"
            )


class ContinuousNonlinearConstraint(ContinuousConstraint, ABC):
    """Abstract base class for continuous nonlinear constraints."""


# Register (un-)structure hooks
converter.register_unstructure_hook(Constraint, unstructure_base)

# Currently affected by a deprecation
# converter.register_structure_hook(Constraint, get_base_structure_hook(Constraint))
converter.register_structure_hook(Constraint, structure_constraints)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Base classes for all meta recommenders."""

import gc
from abc import ABC, abstractmethod
from typing import Any

import cattrs
import pandas as pd
from attrs import define, field

from baybe.objectives.base import Objective
from baybe.recommenders.base import RecommenderProtocol
from baybe.recommenders.pure.base import PureRecommender
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace import SearchSpace
from baybe.serialization import SerialMixin, converter, unstructure_base
from baybe.serialization.core import get_base_structure_hook


@define
class MetaRecommender(SerialMixin, RecommenderProtocol, ABC):
    """Abstract base class for all meta recommenders."""

    _current_recommender: PureRecommender | None = field(default=None, init=False)
    """The current recommender."""

    _used_recommender_ids: set[int] = field(factory=set, init=False)
    """Set of ids from recommenders that were used by this meta recommender."""

    @abstractmethod
    def select_recommender(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> PureRecommender:
        """Select a pure recommender for the given experimentation context.

        See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend` for details
        on the method arguments.
        """

    def get_current_recommender(self) -> PureRecommender:
        """Get the current recommender, if available."""
        if self._current_recommender is None:
            raise RuntimeError(
                f"No recommendation has been requested from the "
                f"'{self.__class__.__name__}' yet. Because the recommender is a "
                f"'{MetaRecommender.__name__}', this means no actual recommender has "
                f"been selected so far. The recommender will be available after the "
                f"next '{self.recommend.__name__}' call."
            )
        return self._current_recommender

    def get_next_recommender(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> PureRecommender:
        """Get the recommender for the next recommendation.

        Returns the next recommender in row that has not yet been used for generating
        recommendations. In case of multiple consecutive calls, this means that
        the same recommender instance is returned until its :meth:`recommend` method
        is called.

        See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend` for details
        on the method arguments.
        """
        # Check if the stored recommender instance can be returned
        if (
            self._current_recommender is not None
            and id(self._current_recommender) not in self._used_recommender_ids
        ):
            recommender = self._current_recommender

        # Otherwise, fetch the next recommender waiting in row
        else:
            recommender = self.select_recommender(
                batch_size=batch_size,
                searchspace=searchspace,
                objective=objective,
                measurements=measurements,
                pending_experiments=pending_experiments,
            )
            self._current_recommender = recommender

        return recommender

    def recommend(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        """See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`."""
        recommender = self.get_next_recommender(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=pending_experiments,
        )

        # Non-predictive recommenders should not be called with an objective or
        # measurements. Using dict value type Any here due to known mypy complication:
        # https://github.com/python/mypy/issues/5382
        optional_args: dict[str, Any] = (
            {}
            if isinstance(recommender, NonPredictiveRecommender)
            else {
                "objective": objective,
                "measurements": measurements,
            }
        )

        recommendations = recommender.recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            pending_experiments=pending_experiments,
            **optional_args,
        )
        self._used_recommender_ids.add(id(recommender))

        return recommendations


# Register (un-)structure hooks
converter.register_unstructure_hook(
    MetaRecommender,
    lambda x: unstructure_base(
        x,
        # TODO: Remove once deprecation got expired:
        overrides=dict(
            allow_repeated_recommendations=cattrs.override(omit=True),
            allow_recommending_already_measured=cattrs.override(omit=True),
        ),
    ),
)
converter.register_structure_hook(
    MetaRecommender, get_base_structure_hook(MetaRecommender)
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Deprecation tests."""

import warnings

import pandas as pd
import pytest

from baybe.acquisition.base import AcquisitionFunction
from baybe.constraints import (
    ContinuousLinearConstraint,
    ContinuousLinearEqualityConstraint,
    ContinuousLinearInequalityConstraint,
)
from baybe.constraints.base import Constraint
from baybe.exceptions import DeprecationError
from baybe.objective import Objective as OldObjective
from baybe.objectives.base import Objective
from baybe.objectives.desirability import DesirabilityObjective
from baybe.parameters.numerical import NumericalContinuousParameter
from baybe.recommenders.pure.bayesian import (
    BotorchRecommender,
    SequentialGreedyRecommender,
)
from baybe.searchspace.continuous import SubspaceContinuous
from baybe.targets.numerical import NumericalTarget


def test_objective_class():
    """Using the deprecated objective class raises a warning."""
    with pytest.warns(DeprecationWarning):
        OldObjective(mode="SINGLE", targets=[NumericalTarget(name="a", mode="MAX")])


deprecated_objective_config = """
{
    "mode": "DESIRABILITY",
    "targets": [
        {
            "type": "NumericalTarget",
            "name": "Yield",
            "mode": "MAX",
            "bounds": [0, 1]
        },
        {
            "type": "NumericalTarget",
            "name": "Waste",
            "mode": "MIN",
            "bounds": [0, 1]
        }
    ],
    "combine_func": "MEAN",
    "weights": [1, 2]
}
"""


def test_objective_config_deserialization():
    """The deprecated objective config format can still be parsed."""
    expected = DesirabilityObjective(
        targets=[
            NumericalTarget("Yield", "MAX", bounds=(0, 1)),
            NumericalTarget("Waste", "MIN", bounds=(0, 1)),
        ],
        scalarizer="MEAN",
        weights=[1, 2],
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        actual = Objective.from_json(deprecated_objective_config)
    assert expected == actual, (expected, actual)


@pytest.mark.parametrize("acqf", ("VarUCB", "qVarUCB"))
def test_acqfs(acqf):
    """Using the deprecated acqf raises a warning."""
    with pytest.warns(DeprecationWarning):
        BotorchRecommender(acquisition_function=acqf)

    with pytest.warns(DeprecationWarning):
        AcquisitionFunction.from_dict({"type": acqf})


def test_acqf_keyword(acqf):
    """Using the deprecated keyword raises an error."""
    with pytest.raises(DeprecationError):
        BotorchRecommender(acquisition_function_cls="qEI")


def test_sequentialgreedyrecommender_class():
    """Using the deprecated `SequentialGreedyRecommender` class raises a warning."""
    with pytest.warns(DeprecationWarning):
        SequentialGreedyRecommender()


def test_samples_random():
    """Using the deprecated `samples_random` method raises a warning."""
    with pytest.warns(DeprecationWarning):
        parameters = [NumericalContinuousParameter("x", (0, 1))]
        SubspaceContinuous(parameters).samples_random(n_points=1)


def test_samples_full_factorial():
    """Using the deprecated `samples_full_factorial` method raises a warning."""
    with pytest.warns(DeprecationWarning):
        parameters = [NumericalContinuousParameter("x", (0, 1))]
        SubspaceContinuous(parameters).samples_full_factorial(n_points=1)


def test_transform_interface(searchspace):
    """Using the deprecated transform interface raises a warning."""
    # Not providing `allow_extra` when there are additional columns
    with pytest.warns(DeprecationWarning):
        searchspace.discrete.transform(
            pd.DataFrame(columns=["additional", *searchspace.discrete.exp_rep.columns])
        )

    # Passing dataframe via `data`
    with pytest.warns(DeprecationWarning):
        searchspace.discrete.transform(
            data=searchspace.discrete.exp_rep, allow_extra=True
        )


def test_surrogate_registration():
    """Using the deprecated registration mechanism raises a warning."""
    from baybe.surrogates import register_custom_architecture

    with pytest.raises(DeprecationError):
        register_custom_architecture()


def test_surrogate_access():
    """Public attribute access to the surrogate model raises a warning."""
    recommender = BotorchRecommender()
    with pytest.warns(DeprecationWarning):
        recommender.surrogate_model


def test_continuous_linear_eq_constraint():
    """Usage of deprecated continuous linear eq constraint raises a warning."""
    with pytest.warns(DeprecationWarning):
        ContinuousLinearEqualityConstraint(["p1", "p2"])


def test_continuous_linear_inq_constraint():
    """Usage of deprecated continuous linear ineq constraint raises a warning."""
    with pytest.warns(DeprecationWarning):
        ContinuousLinearInequalityConstraint(["p1", "p2"])


@pytest.mark.parametrize(
    ("type_", "op"),
    [
        ("ContinuousLinearEqualityConstraint", "="),
        ("ContinuousLinearInequalityConstraint", ">="),
    ],
    ids=["lin_eq", "lin_ineq"],
)
def test_constraint_config_deserialization(type_, op):
    """The deprecated constraint config format can still be parsed."""
    config = """
    {
        "type": "__replace__",
        "parameters": ["p1", "p2", "p3"],
        "coefficients": [1.0, 2.0, 3.0],
        "rhs": 2.0
    }
    """
    config = config.replace("__replace__", type_)

    expected = ContinuousLinearConstraint(
        parameters=["p1", "p2", "p3"],
        operator=op,
        coefficients=[1.0, 2.0, 3.0],
        rhs=2.0,
    )

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        actual = Constraint.from_json(config)
    assert expected == actual, (expected, actual)



## Registering Custom Hooks

# This example demonstrates the basic mechanics of the
# {func}`register_hooks <baybe.utils.basic.register_hooks>` utility,
# which lets you hook into any callable of your choice:
# * We define a hook that is compatible with the general
#   {meth}`RecommenderProtocol.recommend <baybe.recommenders.base.RecommenderProtocol.recommend>`
#   interface,
# * attach it to a recommender,
# * and watch it take action.


### Imports


from dataclasses import dataclass
from time import perf_counter
from types import MethodType

from baybe.parameters import NumericalDiscreteParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.utils import register_hooks

### Defining the Hooks

# We start by defining a simple hook that lets us inspect the names of the parameters
# involved in the recommendation process.
# For this purpose, we match its signature to that of
# {meth}`RecommenderProtocol.recommend <baybe.recommenders.base.RecommenderProtocol.recommend>`:
#

# ```{admonition} Signature components
# :class: note
# Note that you are flexible in designing the signature of your hooks.
# For instance, function parameters and type annotations that you do not need in the
# hook body can simply be omitted.
# The exact rules to follow are described {func}`here <baybe.utils.basic.register_hooks>`.
# ```


def print_parameter_names_hook(self: RandomRecommender, searchspace: SearchSpace):
    """Print the names of the parameters spanning the search space."""
    print(f"Recommender type: {self.__class__.__name__}")
    print(f"Search space parameters: {[p.name for p in searchspace.parameters]}")


# Additionally, we set up a class that provides a combination of hooks for measuring
# the time needed to compute the recommendations:


@dataclass
class ElapsedTimePrinter:
    """Helper class for measuring the time between two calls."""

    last_call_time: float | None = None

    def start(printer_instance):
        """Start the timer."""
        printer_instance.last_call_time = perf_counter()

    def measure(printer_instance, self: RandomRecommender):
        """Measure the elapsed time."""
        if printer_instance.last_call_time is None:
            raise RuntimeError("Must call `start` first!")
        elapsed = perf_counter() - printer_instance.last_call_time
        print(f"Consumed time of {self.__class__.__name__}: {elapsed}")


# ```{admonition} Hook instance vs. target instance
# :class: important
# Notice the difference between the object belonging to the hook-providing class
# (named `printer_instance`) and the object whose method we intend to override
# (named `self`). This distinction is necessary because of
# {ref}`the particular way <BOUND_METHODS>` we attach the hook below, which binds `self`
# to the object carrying the target callable as a method.
# ```


### Monkeypatching

# Next, we create our recommender and monkeypatch its `recommend` method:

timer = ElapsedTimePrinter()
recommender = RandomRecommender()
recommender.recommend = MethodType(
    register_hooks(
        RandomRecommender.recommend,
        pre_hooks=[print_parameter_names_hook, timer.start],
        post_hooks=[timer.measure],
    ),
    recommender,
)

# (BOUND_METHODS)=
# ```{admonition} Bound methods
# :class: important
# Note that the explicit binding via `MethodType` above is required because we
# decorate the (unbound) `RandomRecommender.recommend` **function** with our hooks
# and attach it as an overridden **method** to the recommender instance.
#
# Alternatively, we could have ...
# * ... overridden the class callable itself via
#   `RandomRecommender.recommend = register_hooks(RandomRecommender.recommend, ...)`
#   which, however, would affect all instances of `RandomRecommender` or
# * ... used the bound method of the instance as reference via
#   `recommender.recommend = register_hooks(recommender.recommend, ...)` but then
#   the hooks would not have access to the recommender instance as it is not
#   explicitly exposed in the method's signature.
# ```

### Triggering the Hooks

# When we now apply the recommender in a specific context, we immediately see the
# effect of the hooks:

temperature = NumericalDiscreteParameter("Temperature", values=[90, 105, 120])
concentration = NumericalDiscreteParameter("Concentration", values=[0.057, 0.1, 0.153])
searchspace = SearchSpace.from_product([temperature, concentration])
recommendation = recommender.recommend(batch_size=3, searchspace=searchspace)



"""Utilities for memory usage."""


def bytes_to_human_readable(num: float, /) -> tuple[float, str]:
    """Turn a float number representing a memory byte size into a human-readable format.

    Args:
        num: The number representing a memory size in bytes.

    Returns:
        A tuple with the converted number and its determined human-readable unit.
    """
    for unit in ["bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(num) < 1024.0:
            return num, unit
        num /= 1024.0
    return round(num, 2), "YB"



"""Utilities targeting random number generation."""

import contextlib
import random

import numpy as np


def set_random_seed(seed: int):
    """Set the global random seed.

    Args:
        seed: The chosen global random seed.
    """
    import torch

    # Ensure seed limits
    seed %= 2**32

    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)


@contextlib.contextmanager
def temporary_seed(seed: int):  # noqa: DOC402, DOC404
    """Context manager for setting a temporary random seed.

    Args:
        seed: The chosen random seed.
    """
    import torch

    # Ensure seed limits
    seed %= 2**32

    # Collect the current RNG states
    state_builtin = random.getstate()
    state_np = np.random.get_state()
    state_torch = torch.get_rng_state()

    # Set the requested seed
    set_random_seed(seed)

    # Run the context-specific code
    try:
        yield

    # Restore the original RNG states
    finally:
        random.setstate(state_builtin)
        np.random.set_state(state_np)
        torch.set_rng_state(state_torch)



"""Test if all examples can be run without error."""

import os
import subprocess
from pathlib import Path

import pytest

from baybe._optional.info import CHEM_INSTALLED, ONNX_INSTALLED

# Run these tests in reduced settings
_SMOKE_TEST_CACHE = os.environ.get("SMOKE_TEST", None)
os.environ["SMOKE_TEST"] = "true"


paths = [str(x) for x in Path("examples/").rglob("*.py")]


@pytest.mark.slow
@pytest.mark.skipif(
    not (CHEM_INSTALLED and ONNX_INSTALLED), reason="skipped for core tests"
)
@pytest.mark.parametrize("example", paths, ids=paths)
def test_example(example: str):
    """Test an individual example by running it.

    This runs in a separate process and isolated environment due to problems caused by
    monkeypatching in some examples affecting other tests if they were executed in the
    same environment.
    """
    subprocess.run(["python", example], check=True)


if _SMOKE_TEST_CACHE is not None:
    os.environ["SMOKE_TEST"] = _SMOKE_TEST_CACHE
else:
    os.environ.pop("SMOKE_TEST")



"""Validation tests for search spaces."""

import pandas as pd
import pytest
from pytest import param

from baybe.parameters.numerical import NumericalDiscreteParameter
from baybe.searchspace.validation import get_transform_parameters

parameters = [NumericalDiscreteParameter("d1", [0, 1])]


@pytest.mark.parametrize(
    ("df", "match"),
    [
        param(
            pd.DataFrame(columns=[]),
            r"parameter\(s\) \{'d1'\} cannot be matched",
            id="missing",
        ),
        param(
            pd.DataFrame(columns=["d1", "d2"]),
            r"column\(s\) \{'d2'\} cannot be matched",
            id="extra",
        ),
    ],
)
def test_invalid_transforms(df, match):
    """Transforming dataframes with incorrect columns raises an error."""
    with pytest.raises(ValueError, match=match):
        get_transform_parameters(parameters, df, allow_missing=False, allow_extra=False)


@pytest.mark.parametrize(
    ("df", "missing", "extra"),
    [
        param(pd.DataFrame(columns=["d1"]), False, False, id="equal"),
        param(pd.DataFrame(columns=[]), True, False, id="missing"),
        param(pd.DataFrame(columns=["d1", "d2"]), False, True, id="extra"),
    ],
)
def test_valid_transforms(df, missing, extra):
    """When providing the appropriate flags, the columns of the dataframe to be transformed can be flexibly chosen."""  # noqa
    get_transform_parameters(parameters, df, allow_missing=missing, allow_extra=extra)



"""Functionality for constraint conditions."""

from __future__ import annotations

import gc
import operator as ops
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
from attr import define, field
from attr.validators import in_
from attrs.validators import min_len
from cattrs.gen import override
from funcy import rpartial
from numpy.typing import ArrayLike

from baybe.parameters.validation import validate_unique_values
from baybe.serialization import (
    SerialMixin,
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.utils.numerical import DTypeFloatNumpy

if TYPE_CHECKING:
    import polars as pl


def _is_not_close(x: ArrayLike, y: ArrayLike, rtol: float, atol: float) -> np.ndarray:
    """Return a boolean array indicating where ``x`` and ``y`` are not close.

    The counterpart to ``numpy.isclose``.

    Args:
        x: First input array to compare.
        y: Second input array to compare.
        rtol: The relative tolerance parameter.
        atol: The absolute tolerance parameter.

    Returns:
        A boolean array of where ``x`` and ``y`` are not equal within the
        given tolerances.

    """
    return np.logical_not(_is_close(x, y, rtol=rtol, atol=atol))


def _is_close(x: ArrayLike, y: ArrayLike, rtol: float, atol: float) -> np.ndarray:
    """Return a boolean array indicating where ``x`` and ``y`` are close.

    The equivalent to :func:``numpy.isclose``.
    Using ``numpy.isclose`` with Polars dataframes results in this error:
    ``TypeError: ufunc 'isfinite' not supported for the input types``.

    Args:
        x: First input array to compare.
        y: Second input array to compare.
        rtol: The relative tolerance parameter.
        atol: The absolute tolerance parameter.

    Returns:
        A boolean array of where ``x`` and ``y`` are equal within the
        given tolerances.

    """
    return np.abs(np.subtract(x, y)) <= atol + rtol * np.abs(y)


# provide threshold operators
_threshold_operators: dict[str, Callable] = {
    "<": ops.lt,
    "<=": ops.le,
    "=": rpartial(_is_close, rtol=0.0),
    "==": rpartial(_is_close, rtol=0.0),
    "!=": rpartial(_is_not_close, rtol=0.0),
    ">": ops.gt,
    ">=": ops.ge,
}

# define operators that are eligible for tolerance
_valid_tolerance_operators = ["=", "==", "!="]

_valid_logic_combiners = {
    "AND": ops.and_,
    "OR": ops.or_,
    "XOR": ops.xor,
}


class Condition(ABC, SerialMixin):
    """Abstract base class for all conditions.

    Conditions always evaluate an expression regarding a single parameter.
    Conditions are part of constraints, a constraint can have multiple conditions.
    """

    @abstractmethod
    def evaluate(self, data: pd.Series) -> pd.Series:
        """Evaluate the condition on a given data series.

        Args:
            data: A series containing parameter values.

        Returns:
            A boolean series indicating which elements satisfy the condition.
        """

    @abstractmethod
    def to_polars(self, expr: pl.Expr, /) -> pl.Expr:
        """Apply the condition to a Polars expression.

        Args:
            expr: Input expression, for instance column selection etc.

        Returns:
            An expression that can be used for filtering.
        """


@define
class ThresholdCondition(Condition):
    """Class for modelling threshold-based conditions."""

    # object variables
    threshold: float = field()
    """The threshold value used in the condition."""

    operator: str = field(validator=[in_(_threshold_operators)])
    """The operator used in the condition."""

    tolerance: float | None = field()
    """A numerical tolerance. Set to a reasonable default tolerance."""

    @tolerance.default
    def _tolerance_default(self) -> float | None:
        """Create the default value for the tolerance."""
        # Default value for the tolerance.
        return 1e-8 if self.operator in _valid_tolerance_operators else None

    @tolerance.validator
    def _validate_tolerance(self, _: Any, value: float) -> None:  # noqa: DOC101, DOC103
        """Validate the threshold condition tolerance.

        Raises:
            ValueError: If the operator does not allow for setting a tolerance.
            ValueError: If the operator allows for setting a tolerance, but the provided
                tolerance is either less than 0 or ``None``.
        """
        if (self.operator not in _valid_tolerance_operators) and (value is not None):
            raise ValueError(
                f"Setting the tolerance for a threshold condition is only valid "
                f"with the following operators: {_valid_tolerance_operators}."
            )
        if self.operator in _valid_tolerance_operators:
            if (value is None) or (value <= 0.0):
                raise ValueError(
                    f"When using a tolerance-enabled operator"
                    f" ({_valid_tolerance_operators}) the tolerance cannot be None "
                    f"or <= 0.0, but was {value}."
                )

    def _make_operator_function(self):
        """Generate a function using operators to filter out undesired rows."""
        func = rpartial(_threshold_operators[self.operator], self.threshold)
        if self.operator in _valid_tolerance_operators:
            func = rpartial(func, atol=self.tolerance)
        return func

    def evaluate(self, data: pd.Series) -> pd.Series:  # noqa: D102
        # See base class.
        if data.dtype.kind not in "iufb":
            raise ValueError(
                "You tried to apply a threshold condition to non-numeric data. "
                "This operation is error-prone and not supported. Only use threshold "
                "conditions with numerical parameters."
            )
        func = self._make_operator_function()
        return data.apply(func)

    def to_polars(self, expr: pl.Expr, /) -> pl.Expr:  # noqa: D102
        # See base class.
        op = self._make_operator_function()
        return op(expr)


@define
class SubSelectionCondition(Condition):
    """Class for defining valid parameter entries."""

    # object variables
    _selection: tuple = field(
        converter=tuple,
        # FIXME[typing]: https://github.com/python-attrs/attrs/issues/1197
        validator=[
            min_len(1),
            validate_unique_values,  # type: ignore
        ],
    )
    """The internal list of items which are considered valid."""

    @property
    def selection(self) -> tuple:  # noqa: D102
        """The list of items which are considered valid."""
        return tuple(
            DTypeFloatNumpy(itm) if isinstance(itm, (float, int, bool)) else itm
            for itm in self._selection
        )

    def evaluate(self, data: pd.Series) -> pd.Series:  # noqa: D102
        # See base class.
        return data.isin(self.selection)

    def to_polars(self, expr: pl.Expr, /) -> pl.Expr:  # noqa: D102
        # See base class.
        return expr.is_in(self.selection)


# Register (un-)structure hooks
_overrides = {
    "_selection": override(rename="selection"),
}
# FIXME[typing]: https://github.com/python/mypy/issues/4717
converter.register_structure_hook(
    Condition,
    get_base_structure_hook(Condition, overrides=_overrides),  # type: ignore
)
converter.register_unstructure_hook(
    Condition, partial(unstructure_base, overrides=_overrides)
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Random forest surrogates."""

from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING, Any, ClassVar, Protocol

import numpy as np
from attr import define, field
from sklearn.ensemble import RandomForestRegressor

from baybe.parameters.base import Parameter
from baybe.surrogates.base import Surrogate
from baybe.surrogates.utils import batchify_ensemble_predictor, catch_constant_targets
from baybe.surrogates.validation import get_model_params_validator
from baybe.utils.plotting import to_string

if TYPE_CHECKING:
    from botorch.models.ensemble import EnsemblePosterior
    from botorch.models.transforms.input import InputTransform
    from botorch.models.transforms.outcome import OutcomeTransform
    from torch import Tensor


class _Predictor(Protocol):
    """A basic predictor."""

    def predict(self, x: np.ndarray, /) -> np.ndarray: ...


@catch_constant_targets
@define
class RandomForestSurrogate(Surrogate):
    """A random forest surrogate model."""

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    model_params: dict[str, Any] = field(
        factory=dict,
        converter=dict,
        validator=get_model_params_validator(RandomForestRegressor.__init__),
    )
    """Optional model parameter that will be passed to the surrogate constructor."""

    _model: RandomForestRegressor | None = field(init=False, default=None, eq=False)
    """The actual model."""

    @staticmethod
    def _make_parameter_scaler_factory(
        parameter: Parameter,
    ) -> type[InputTransform] | None:
        # See base class.

        # Tree-like models do not require any input scaling
        return None

    @staticmethod
    def _make_target_scaler_factory() -> type[OutcomeTransform] | None:
        # See base class.

        # Tree-like models do not require any output scaling
        return None

    def _posterior(self, candidates_comp_scaled: Tensor, /) -> EnsemblePosterior:
        # See base class.

        from botorch.models.ensemble import EnsemblePosterior

        @batchify_ensemble_predictor
        def predict(candidates_comp_scaled: Tensor) -> Tensor:
            """Make the end-to-end ensemble prediction."""
            import torch

            # FIXME[typing]: It seems there is currently no better way to inform the
            #   type checker that the attribute is available at the time of the
            #   function call
            assert self._model is not None

            return torch.from_numpy(
                self._predict_ensemble(
                    self._model.estimators_, candidates_comp_scaled.numpy()
                )
            )

        return EnsemblePosterior(predict(candidates_comp_scaled).unsqueeze(-1))

    @staticmethod
    def _predict_ensemble(
        predictors: Collection[_Predictor], candidates: np.ndarray
    ) -> np.ndarray:
        """Evaluate an ensemble of predictors on a given candidate set."""
        # Extract shapes
        n_candidates = len(candidates)
        n_estimators = len(predictors)

        # Evaluate all trees
        predictions = np.zeros((n_estimators, n_candidates))
        for p, predictor in enumerate(predictors):
            predictions[p] = predictor.predict(candidates)

        return predictions

    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        # See base class.
        self._model = RandomForestRegressor(**(self.model_params))
        self._model.fit(train_x.numpy(), train_y.numpy().ravel())

    def __str__(self) -> str:
        fields = [to_string("Model Params", self.model_params, single_line=True)]
        return to_string(super().__str__(), *fields)



"""Test for imposing continuous constraints."""

import numpy as np
import pytest
from pytest import param

from baybe.constraints import ContinuousLinearConstraint
from tests.conftest import run_iterations


@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_1"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
def test_equality1(campaign, n_iterations, batch_size):
    """Test equality constraint with equal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert np.allclose(1.0 * res["Conti_finite1"] + 1.0 * res["Conti_finite2"], 0.3)


@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_2"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
def test_equality2(campaign, n_iterations, batch_size):
    """Test equality constraint with unequal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert np.allclose(1.0 * res["Conti_finite1"] + 3.0 * res["Conti_finite2"], 0.3)


@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_3"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
def test_inequality1(campaign, n_iterations, batch_size):
    """Test inequality constraint with equal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert (1.0 * res["Conti_finite1"] + 1.0 * res["Conti_finite2"]).ge(0.299).all()


@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_4"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
def test_inequality2(campaign, n_iterations, batch_size):
    """Test inequality constraint with unequal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert (1.0 * res["Conti_finite1"] + 3.0 * res["Conti_finite2"]).ge(0.299).all()


@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_6"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
def test_inequality3(campaign, n_iterations, batch_size):
    """Test inequality constraint with unequal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert (1.0 * res["Conti_finite1"] + 3.0 * res["Conti_finite2"]).le(0.301).all()


@pytest.mark.slow
@pytest.mark.parametrize(
    "parameter_names",
    [["Solvent_1", "Conti_finite1", "Conti_finite3", "Conti_finite2"]],
)
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_1"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
@pytest.mark.parametrize("n_grid_points", [5], ids=["grid5"])
def test_hybridspace_eq(campaign, n_iterations, batch_size):
    """Test equality constraint with equal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert np.allclose(1.0 * res["Conti_finite1"] + 1.0 * res["Conti_finite2"], 0.3)


@pytest.mark.slow
@pytest.mark.parametrize(
    "parameter_names",
    [["Solvent_1", "Conti_finite1", "Conti_finite3", "Conti_finite2"]],
)
@pytest.mark.parametrize("constraint_names", [["ContiConstraint_3"]])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
@pytest.mark.parametrize("n_grid_points", [5], ids=["grid5"])
def test_hybridspace_ineq(campaign, n_iterations, batch_size):
    """Test inequality constraint with equal weights."""
    run_iterations(campaign, n_iterations, batch_size, add_noise=False)
    res = campaign.measurements
    print(res)

    assert (1.0 * res["Conti_finite1"] + 1.0 * res["Conti_finite2"]).ge(0.299).all()


@pytest.mark.parametrize(
    ("parameters", "coefficients", "rhs", "op"),
    [
        param(["A", "B"], [1.0], 0.0, "=", id="eq_too_few_coeffs"),
        param(["A", "B"], [1.0, 2.0, 3.0], 0.0, "=", id="eq_too_many_coeffs"),
        param(["A", "B"], [1.0, 2.0], "bla", "=", id="eq_invalid_rhs"),
        param(["A", "B"], [1.0], 0.0, ">=", id="ineq_too_few_coeffs"),
        param(["A", "B"], [1.0, 2.0, 3.0], 0.0, ">=", id="ineq_too_many_coeffs"),
        param(["A", "B"], [1.0, 2.0], "bla", ">=", id="ineq_invalid_rhs"),
        param(["A", "B"], [1.0, 2.0], 0.0, "invalid", id="ineq_invalid_operator1"),
        param(["A", "B"], [1.0, 2.0], 0.0, 2.0, id="ineq_invalid_operator1"),
    ],
)
def test_invalid_constraints(parameters, coefficients, rhs, op):
    """Test invalid continuous constraint creations."""
    with pytest.raises(ValueError):
        ContinuousLinearConstraint(
            parameters=parameters, operator=op, coefficients=coefficients, rhs=rhs
        )



"""# Surrogate Models

The purpose of this script is to compare different surrogate models and verify that
their predictions are invariant to changes in location/scale of the underlying target
function.

This means that the displayed function approximation should always look the same when
the input and output locations/scales are changed.
"""  # noqa: D415

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import streamlit as st
import torch
from funcy import rpartial

from baybe.acquisition.acqfs import qLogExpectedImprovement
from baybe.acquisition.base import AcquisitionFunction
from baybe.parameters.numerical import NumericalDiscreteParameter
from baybe.recommenders.pure.bayesian.botorch import BotorchRecommender
from baybe.searchspace import SearchSpace
from baybe.surrogates import CustomONNXSurrogate
from baybe.surrogates.base import Surrogate
from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate
from baybe.targets.numerical import NumericalTarget
from baybe.utils.basic import get_subclasses
from baybe.utils.random import set_random_seed

# Number of values used for the input parameter
N_PARAMETER_VALUES = 200


def cubic(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Cubic test function."""
    out = amplitude * np.power((x - x_min) / (x_max - x_min), 3) + bias
    return out


def sin(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Sinusoid test function."""
    out = amplitude * np.sin((x - x_min) / (x_max - x_min) * 2 * np.pi) + bias
    return out


def constant(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Constant test function."""
    out = np.full(x.shape, fill_value=bias)
    return out


def linear(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Linear test function."""
    out = amplitude * x + bias
    return out


def main():
    """Create the streamlit dashboard."""
    # Show module docstring in dashboard
    st.info(__doc__)

    # Collect all available test functions
    test_functions = {
        "Sine": sin,
        "Constant": constant,
        "Linear": linear,
        "Cubic": cubic,
    }

    # Collect all available surrogate models
    surrogate_model_classes = {
        cls.__name__: cls
        for cls in get_subclasses(Surrogate)
        if not issubclass(cls, CustomONNXSurrogate)
    }
    surrogate_model_names = list(surrogate_model_classes.keys())

    # Collect all available acquisition functions
    acquisition_function_classes = {
        cls.__name__: cls for cls in get_subclasses(AcquisitionFunction)
    }
    acquisition_function_names = list(acquisition_function_classes.keys())

    # Streamlit simulation parameters
    st.sidebar.markdown("# Domain")
    st_random_seed = int(st.sidebar.number_input("Random seed", value=1337))
    st_function_name = st.sidebar.selectbox(
        "Test function", list(test_functions.keys())
    )
    st_target_mode = st.sidebar.radio(
        "Objective",
        ["MAX", "MIN"],
        format_func=lambda x: {"MAX": "Maximization", "MIN": "Minimization"}[x],
        horizontal=True,
    )
    st.sidebar.markdown("---")
    st.sidebar.markdown("# Model")
    st_surrogate_name = st.sidebar.selectbox(
        "Surrogate model",
        surrogate_model_names,
        surrogate_model_names.index(GaussianProcessSurrogate.__name__),
    )
    st_acqf_name = st.sidebar.selectbox(
        "Acquisition function",
        acquisition_function_names,
        acquisition_function_names.index(qLogExpectedImprovement.__name__),
    )
    st_n_training_points = st.sidebar.slider("Number of training points", 1, 20, 5)
    st_n_recommendations = st.sidebar.slider("Number of recommendations", 1, 20, 5)
    st.sidebar.markdown("---")
    st.sidebar.markdown("# Validation")
    st.sidebar.markdown(
        """
        When scaling is implemented correctly, the plot should remain static (except for
        the axis labels) when changing the following parameters:
        """
    )
    st_upper_parameter_limit = st.sidebar.slider(
        "Upper parameter limit", 0.0, 100.0, 1.0
    )
    st_lower_parameter_limit = st.sidebar.slider(
        "Lower parameter limit", -100.0, 0.0, 0.0
    )
    st_function_amplitude = st.sidebar.slider("Function amplitude", 1.0, 100.0, 1.0)
    st_function_bias = st.sidebar.slider("Function bias", -100.0, 100.0, 0.0)

    # Set the chosen random seed
    set_random_seed(st_random_seed)

    # Construct the specific test function
    fun = rpartial(
        test_functions[st_function_name],
        st_lower_parameter_limit,
        st_upper_parameter_limit,
        st_function_amplitude,
        st_function_bias,
    )

    # Create the training data
    train_x = np.random.uniform(
        st_lower_parameter_limit, st_upper_parameter_limit, st_n_training_points
    )
    train_y = fun(train_x)
    measurements = pd.DataFrame({"x": train_x, "y": train_y})

    # Create the plotting grid and corresponding target values
    test_x = np.linspace(
        st_lower_parameter_limit, st_upper_parameter_limit, N_PARAMETER_VALUES
    )
    test_y = fun(test_x)
    candidates = pd.DataFrame({"x": test_x, "y": test_y})

    # Create the searchspace and objective
    parameter = NumericalDiscreteParameter(
        name="x",
        values=np.linspace(
            st_lower_parameter_limit, st_upper_parameter_limit, N_PARAMETER_VALUES
        ),
    )
    searchspace = SearchSpace.from_product(parameters=[parameter])
    objective = NumericalTarget(name="y", mode=st_target_mode).to_objective()

    # Create the surrogate model, acquisition function, and the recommender
    surrogate_model = surrogate_model_classes[st_surrogate_name]()
    acqf = acquisition_function_classes[st_acqf_name]()
    recommender = BotorchRecommender(
        surrogate_model=surrogate_model, acquisition_function=acqf
    )

    # Get the recommendations and extract the posterior mean / standard deviation
    recommendations = recommender.recommend(
        st_n_recommendations, searchspace, objective, measurements
    )
    with torch.no_grad():
        posterior = surrogate_model.posterior(candidates)
    mean = posterior.mean.squeeze().numpy()
    std = posterior.variance.sqrt().squeeze().numpy()

    # Visualize the test function, training points, model predictions, recommendations
    fig = plt.figure()
    plt.plot(test_x, test_y, color="tab:blue", label="Test function")
    plt.plot(train_x, train_y, "o", color="tab:blue")
    plt.plot(test_x, mean, color="tab:red", label="Surrogate model")
    plt.fill_between(test_x, mean - std, mean + std, alpha=0.2, color="tab:red")
    plt.vlines(
        recommendations, *plt.gca().get_ylim(), color="k", label="Recommendations"
    )
    plt.legend()
    st.pyplot(fig)


if __name__ == "__main__":
    main()



"""Kernel tests."""

from typing import Any

import numpy as np
import torch
from attrs import asdict, has
from hypothesis import given

from baybe.kernels.base import BasicKernel, Kernel

from .hypothesis_strategies.kernels import kernels

# TODO: Consider deprecating these attribute names to avoid inconsistencies
_RENAME_DICT: dict[str, str] = {
    "base_kernels": "kernels",
}
"""Dictionary for resolving name differences between BayBE and GPyTorch attributes."""


def validate_gpytorch_kernel_components(obj: Any, mapped: Any, **kwargs) -> None:
    """Validate that all kernel components are correctly translated to GPyTorch.

    Args:
        obj: An object occurring as part of a BayBE kernel.
        mapped: The corresponding object in the translated GPyTorch kernel.
        **kwargs: Optional kernel arguments that were passed to the GPyTorch kernel.
    """
    # Assert that the kernel kwargs are correctly mapped
    if isinstance(obj, BasicKernel):
        for k, v in kwargs.items():
            assert torch.tensor(getattr(mapped, k)).equal(torch.tensor(v))

    # Compare attribute by attribute
    for name, component in asdict(obj, recurse=False).items():
        # If the BayBE component is `None`, the GPyTorch component might not even
        # exist, so we skip
        if component is None:
            continue

        # Resolve attribute naming differences
        mapped_name = _RENAME_DICT.get(name, name)

        # If the attribute does not exist in the GPyTorch version, it must be an
        # initial value. Because setting initial values involves going through
        # constraint transformations on GPyTorch side (i.e., difference between
        # `<attr>` and `raw_<attr>`), the numerical values will not be exact, so
        # we check only for approximate matches.
        if (mapped_component := getattr(mapped, mapped_name, None)) is None:
            assert name.endswith("_initial_value")
            assert np.allclose(
                component,
                getattr(mapped, name.removesuffix("_initial_value")).detach().numpy(),
            )

        # If the component is itself another attrs object, recurse
        elif has(component):
            validate_gpytorch_kernel_components(component, mapped_component, **kwargs)

        # Same for collections of BayBE objects (coming from composite kernels)
        elif isinstance(component, tuple) and all(has(c) for c in component):
            for c, m in zip(component, mapped_component):
                validate_gpytorch_kernel_components(c, m, **kwargs)

        # On the lowest component level, simply check for equality
        else:
            assert component == mapped_component


@given(kernels())
def test_kernel_assembly(kernel: Kernel):
    """Turning a BayBE kernel into a GPyTorch kernel raises no errors and all its
    components are translated correctly."""  # noqa
    # Create some arbitrary kernel kwargs to ensure that they are correctly translated
    kwargs = dict(
        ard_num_dims=np.random.randint(0, 32),
        batch_shape=torch.Size(
            [np.random.randint(0, 4) for _ in range(np.random.randint(0, 4))]
        ),
        active_dims=[np.random.randint(0, 4) for _ in range(np.random.randint(0, 4))],
    )

    k = kernel.to_gpytorch(**kwargs)
    validate_gpytorch_kernel_components(kernel, k, **kwargs)



"""Base class for all Bayesian recommenders."""

import gc
import warnings
from abc import ABC

import pandas as pd
from attrs import define, field, fields

from baybe.acquisition.acqfs import qLogExpectedImprovement
from baybe.acquisition.base import AcquisitionFunction
from baybe.acquisition.utils import convert_acqf
from baybe.exceptions import DeprecationError, InvalidSurrogateModelError
from baybe.objectives.base import Objective
from baybe.recommenders.pure.base import PureRecommender
from baybe.searchspace import SearchSpace
from baybe.surrogates import CustomONNXSurrogate, GaussianProcessSurrogate
from baybe.surrogates.base import IndependentGaussianSurrogate, SurrogateProtocol


@define
class BayesianRecommender(PureRecommender, ABC):
    """An abstract class for Bayesian Recommenders."""

    _surrogate_model: SurrogateProtocol = field(
        alias="surrogate_model", factory=GaussianProcessSurrogate
    )
    """The used surrogate model."""

    acquisition_function: AcquisitionFunction = field(
        converter=convert_acqf, factory=qLogExpectedImprovement
    )
    """The used acquisition function class."""

    _botorch_acqf = field(default=None, init=False)
    """The current acquisition function."""

    acquisition_function_cls: str | None = field(default=None, kw_only=True)
    "Deprecated! Raises an error when used."

    @acquisition_function_cls.validator
    def _validate_deprecated_argument(self, _, value) -> None:
        """Raise DeprecationError if old acquisition_function_cls parameter is used."""
        if value is not None:
            raise DeprecationError(
                "Passing 'acquisition_function_cls' to the constructor is deprecated. "
                "The parameter has been renamed to 'acquisition_function'."
            )

    @property
    def surrogate_model(self) -> SurrogateProtocol:
        """Deprecated!"""
        warnings.warn(
            f"Accessing the surrogate model via 'surrogate_model' has been "
            f"deprecated. Use '{self.get_surrogate.__name__}' instead to get the "
            f"trained model instance (or "
            f"'{fields(type(self))._surrogate_model.name}' to access the raw object).",
            DeprecationWarning,
        )
        return self._surrogate_model

    def get_surrogate(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
    ) -> SurrogateProtocol:
        """Get the trained surrogate model."""
        # This fit applies internal caching and does not necessarily involve computation
        self._surrogate_model.fit(searchspace, objective, measurements)

        return self._surrogate_model

    def _setup_botorch_acqf(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
    ) -> None:
        """Create the acquisition function for the current training data."""  # noqa: E501
        surrogate = self.get_surrogate(searchspace, objective, measurements)
        self._botorch_acqf = self.acquisition_function.to_botorch(
            surrogate,
            searchspace,
            objective,
            measurements,
            pending_experiments,
        )

    def recommend(  # noqa: D102
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        # See base class.

        if objective is None:
            raise NotImplementedError(
                f"Recommenders of type '{BayesianRecommender.__name__}' require "
                f"that an objective is specified."
            )

        if (measurements is None) or (len(measurements) == 0):
            raise NotImplementedError(
                f"Recommenders of type '{BayesianRecommender.__name__}' do not support "
                f"empty training data."
            )

        if (
            isinstance(self._surrogate_model, IndependentGaussianSurrogate)
            and batch_size > 1
        ):
            raise InvalidSurrogateModelError(
                f"The specified surrogate model of type "
                f"'{self._surrogate_model.__class__.__name__}' "
                f"cannot be used for batch recommendation."
            )

        if isinstance(self._surrogate_model, CustomONNXSurrogate):
            CustomONNXSurrogate.validate_compatibility(searchspace)

        self._setup_botorch_acqf(
            searchspace, objective, measurements, pending_experiments
        )

        return super().recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=pending_experiments,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Tests features of the Campaign object."""

import pytest
from pytest import param

from .conftest import run_iterations


@pytest.mark.parametrize(
    "target_names",
    [
        param(["Target_max"], id="max"),
        param(["Target_min"], id="min"),
        param(["Target_max_bounded"], id="max_b"),
        param(["Target_min_bounded"], id="min_b"),
        param(["Target_match_bell"], id="match_bell"),
        param(["Target_match_triangular"], id="match_tri"),
        param(
            ["Target_max_bounded", "Target_min_bounded", "Target_match_triangular"],
            id="desirability",
        ),
    ],
)
@pytest.mark.parametrize("batch_size", [2], ids=["b2"])
@pytest.mark.parametrize("n_iterations", [2], ids=["i2"])
def test_get_surrogate(campaign, n_iterations, batch_size):
    """Test successful extraction of the surrogate model."""
    run_iterations(campaign, n_iterations, batch_size)

    model = campaign.get_surrogate()
    assert model is not None, "Something went wrong during surrogate model extraction."
